from __future__ import annotations

import inspect
import logging
from types import ModuleType
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd

from great_expectations.compatibility import aws, pydantic, pyspark, trino
from great_expectations.compatibility.bigquery import (
    BIGQUERY_GEO_SUPPORT,
    bigquery_types_tuple,
)
from great_expectations.compatibility.bigquery import (
    sqlalchemy_bigquery as BigQueryDialect,
)
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.execution_engine.sqlalchemy_dialect import (
    GXSqlDialect,  # noqa: TC001, RUF100 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import COLUMN_DESCRIPTION
from great_expectations.expectations.registry import get_metric_kwargs
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from great_expectations.util import (
    get_clickhouse_sqlalchemy_potential_type,
    get_pyathena_potential_type,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.core import ExpectationValidationResult
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs
    from great_expectations.validator.validator import ValidationDependencies

logger = logging.getLogger(__name__)

try:
    import teradatasqlalchemy.dialect
    import teradatasqlalchemy.types as teradatatypes
except ImportError:
    teradatasqlalchemy = None

try:
    import clickhouse_sqlalchemy
    import clickhouse_sqlalchemy.types as ch_types
except (ImportError, KeyError):
    clickhouse_sqlalchemy = None
    ch_types = None

EXPECTATION_SHORT_DESCRIPTION = "Expect a column to contain values of a specified data type."
TYPE__DESCRIPTION = """
    A string representing the data type that each column should have as entries. \
    Valid types are defined by the current backend implementation and are dynamically loaded.
    """
DATA_QUALITY_ISSUES = [DataQualityIssues.SCHEMA.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnValuesToBeOfType(ColumnMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnValuesToBeOfType is a \
    Column Map Expectation \
    for typed-column backends, and also for Pandas Datasources where the column dtype and provided \
    type_ are unambiguous constraints (any dtype except 'object' or dtype of 'object' with \
    type_ specified as 'object').

    For Pandas columns with dtype of 'object' ExpectColumnValuesToBeOfType will
    independently check each row's type.

    Column Map Expectations are one of the most common types of Expectation.
    They are evaluated for a single column and ask a yes/no question for every row in that column.
    Based on the result, they then calculate the percentage of rows that gave a positive answer. If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        type\\_ (str): \
            {TYPE__DESCRIPTION}
            For example, valid types for Pandas Datasources include any numpy dtype values \
            (such as 'int64') or native python types (such as 'int'), whereas valid types \
            for a SqlAlchemy Datasource include types named by the current driver such as 'INTEGER' \
            in most SQL dialects and 'TEXT' in dialects such as postgresql. \
            Valid types for Spark Datasources include 'StringType', 'BooleanType' and other \
            pyspark-defined type names. Note that the strings representing these \
            types are sometimes case-sensitive. For instance, with a Pandas backend `timestamp` \
            will be unrecognized and fail the expectation, while `Timestamp` would pass with valid data.

    Other Parameters:
        mostly (None or a float between 0 and 1): \
            Successful if at least mostly fraction of values match the expectation. \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See also:
        [ExpectColumnValuesToBeInTypeList](https://greatexpectations.io/expectations/expect_column_values_to_be_in_type_list)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	"12345" 1
            1 	"abcde" 2
            2 	"1b3d5" 3

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnValuesToBeOfType(
                    column="test2",
                    type_="NUMBER"
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 0,
                    "unexpected_percent": 0.0,
                    "partial_unexpected_list": [],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 0.0,
                    "unexpected_percent_nonmissing": 0.0
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnValuesToBeOfType(
                    column="test",
                    type_="DOUBLE"
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 3,
                    "unexpected_percent": 100.0,
                    "partial_unexpected_list": [
                        "12345",
                        "abcde",
                        "1b3d5"
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 100.0,
                    "unexpected_percent_nonmissing": 100.0
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    type_: Union[str, SuiteParameterDict] = pydantic.Field(description=TYPE__DESCRIPTION)

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "column_values.of_type"
    domain_keys: ClassVar[Tuple[str, ...]] = (
        "column",
        "row_condition",
        "condition_parser",
    )
    success_keys = (
        "type_",
        "mostly",
    )
    args_keys = (
        "column",
        "type_",
    )

    class Config:
        title = "Expect column values to be of type"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[ExpectColumnValuesToBeOfType]) -> None:
            ColumnMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @override
    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("type_", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if params.mostly and params.mostly.value < 1.0:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            template_str = "values must be of type $type_, at least $mostly_pct % of the time."
        else:
            template_str = "values must be of type $type_."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @override
    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")

        kwargs = configuration.kwargs if configuration is not None else {}

        params = substitute_none_for_missing(
            kwargs,
            ["column", "type_", "mostly", "row_condition", "condition_parser"],
        )

        if params["mostly"] is not None:
            if isinstance(params["mostly"], (int, float)) and params["mostly"] < 1.0:
                params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
                # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")  # noqa: E501 # FIXME CoP
            template_str = "values must be of type $type_, at least $mostly_pct % of the time."
        else:
            template_str = "values must be of type $type_."

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    def _validate_pandas(
        self,
        actual_column_type,
        expected_type,
    ):
        if expected_type is None:
            success = True
        else:
            comp_types = []

            try:
                comp_types.append(np.dtype(expected_type).type)
            except TypeError:
                try:
                    pd_type = getattr(pd, expected_type)
                    if isinstance(pd_type, type):
                        comp_types.append(pd_type)
                except AttributeError:
                    pass

                try:
                    pd_type = getattr(pd.core.dtypes.dtypes, expected_type)
                    if isinstance(pd_type, type):
                        comp_types.append(pd_type)
                except AttributeError:
                    pass

            native_type = _native_type_type_map(expected_type)
            if native_type is not None:
                comp_types.extend(native_type)

            success = actual_column_type.type in comp_types

        return {
            "success": success,
            "result": {"observed_value": actual_column_type.type.__name__},
        }

    def _validate_sqlalchemy(self, actual_column_type, expected_type, execution_engine):
        # Our goal is to be as explicit as possible. We will match the dialect
        # if that is possible. If there is no dialect available, we *will*
        # match against a top-level SqlAlchemy type.
        #
        # This is intended to be a conservative approach.
        #
        # In particular, we *exclude* types that would be valid under an ORM
        # such as "float" for postgresql with this approach

        if expected_type is None:
            success = True
        elif execution_engine.dialect_name in [
            GXSqlDialect.DATABRICKS,
            GXSqlDialect.POSTGRESQL,
            GXSqlDialect.SNOWFLAKE,
        ]:
            # For these dialects, actual_column_type should be a string or CaseInsensitiveString
            if isinstance(actual_column_type, str):
                # CaseInsensitiveString objects will automatically do case-insensitive comparison
                success = actual_column_type == expected_type
            else:
                # Handle the case where it's not a string type
                # This should never happen, but we'll handle it just in case
                # the column type should be converted to a CaseInsensitiveString
                # for these three dialects in metrics/util.py:get_sqlalchemy_column_metadata
                success = str(actual_column_type).lower() == expected_type.lower()

            return {
                "success": success,
                "result": {"observed_value": actual_column_type},
            }
        else:
            types = _get_potential_sqlalchemy_types(
                execution_engine=execution_engine, expected_type=expected_type
            )
            success = isinstance(actual_column_type, tuple(types))

        return {
            "success": success,
            "result": {"observed_value": type(actual_column_type).__name__},
        }

    def _validate_spark(
        self,
        actual_column_type,
        expected_type,
    ):
        if expected_type is None:
            success = True
        else:
            types = []
            try:
                type_class = getattr(pyspark.types, expected_type)
                types.append(type_class)
            except AttributeError:
                logger.debug(f"Unrecognized type: {expected_type}")
            if len(types) == 0:
                raise ValueError("No recognized spark types in expected_types_list")  # noqa: TRY003 # FIXME CoP
            types = tuple(types)
            success = isinstance(actual_column_type, types)
        return {
            "success": success,
            "result": {"observed_value": type(actual_column_type).__name__},
        }

    @override
    def get_validation_dependencies(
        self,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> ValidationDependencies:
        from great_expectations.execution_engine import (
            PandasExecutionEngine,
        )

        # This calls BatchExpectation.get_validation_dependencies to set baseline validation_dependencies for the aggregate version  # noqa: E501 # FIXME CoP
        # of the expectation.
        # We need to keep this as super(ColumnMapExpectation, self), which calls
        # BatchExpectation.get_validation_dependencies instead of ColumnMapExpectation.get_validation_dependencies.  # noqa: E501 # FIXME CoP
        # This is because the map version of this expectation is only supported for Pandas, so we want the aggregate  # noqa: E501 # FIXME CoP
        # version for the other backends.
        validation_dependencies: ValidationDependencies = super(
            ColumnMapExpectation, self
        ).get_validation_dependencies(execution_engine, runtime_configuration)

        configuration = self.configuration

        # Only PandasExecutionEngine supports the column map version of the expectation.
        kwargs = configuration.kwargs if configuration else {}

        if isinstance(execution_engine, PandasExecutionEngine):
            column_name = kwargs.get("column")
            expected_type = kwargs.get("type_")
            metric_kwargs = get_metric_kwargs(
                metric_name="table.column_types",
                configuration=configuration,
                runtime_configuration=runtime_configuration,
            )
            metric_domain_kwargs = metric_kwargs.get("metric_domain_kwargs", {})
            metric_value_kwargs = metric_kwargs.get("metric_value_kwargs")
            table_column_types_configuration = MetricConfiguration(
                "table.column_types",
                metric_domain_kwargs=metric_domain_kwargs,
                metric_value_kwargs=metric_value_kwargs,
            )
            actual_column_types_list = execution_engine.resolve_metrics(
                [table_column_types_configuration]
            )[table_column_types_configuration.id]
            try:
                actual_column_type = [
                    type_dict["type"]
                    for type_dict in actual_column_types_list
                    if type_dict["name"] == column_name
                ][0]
            except IndexError:
                actual_column_type = None

            # only use column map version if column dtype is object
            if (
                actual_column_type
                and actual_column_type.type.__name__ == "object_"
                and expected_type
                not in [
                    "object",
                    "object_",
                    "O",
                    None,
                ]
            ):
                # this resets validation_dependencies using  ColumnMapExpectation.get_validation_dependencies  # noqa: E501 # FIXME CoP
                validation_dependencies = super().get_validation_dependencies(
                    execution_engine, runtime_configuration
                )

        # this adds table.column_types dependency for both aggregate and map versions of expectation
        column_types_metric_kwargs = get_metric_kwargs(
            metric_name="table.column_types",
            configuration=configuration,
            runtime_configuration=runtime_configuration,
        )
        validation_dependencies.set_metric_configuration(
            metric_name="table.column_types",
            metric_configuration=MetricConfiguration(
                metric_name="table.column_types",
                metric_domain_kwargs=column_types_metric_kwargs["metric_domain_kwargs"],
                metric_value_kwargs=column_types_metric_kwargs["metric_value_kwargs"],
            ),
        )

        return validation_dependencies

    @override
    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        from great_expectations.execution_engine import (
            PandasExecutionEngine,
            SparkDFExecutionEngine,
            SqlAlchemyExecutionEngine,
        )

        configuration = self.configuration
        column_name = configuration.kwargs.get("column")
        expected_type = configuration.kwargs.get("type_")
        actual_column_types_list = metrics.get("table.column_types", [])
        actual_column_type = [
            type_dict["type"]
            for type_dict in actual_column_types_list
            if type_dict["name"] == column_name
        ][0]

        if isinstance(execution_engine, PandasExecutionEngine):
            # only PandasExecutionEngine supports map version of expectation and
            # only when column type is object
            if actual_column_type.type.__name__ == "object_" and expected_type not in [
                "object",
                "object_",
                "O",
                None,
            ]:
                # this calls ColumnMapMetric._validate
                return super()._validate(metrics, runtime_configuration, execution_engine)
            return self._validate_pandas(
                actual_column_type=actual_column_type, expected_type=expected_type
            )
        elif isinstance(execution_engine, SqlAlchemyExecutionEngine):
            return self._validate_sqlalchemy(
                actual_column_type=actual_column_type,
                expected_type=expected_type,
                execution_engine=execution_engine,
            )
        elif isinstance(execution_engine, SparkDFExecutionEngine):
            return self._validate_spark(
                actual_column_type=actual_column_type, expected_type=expected_type
            )


def _get_potential_sqlalchemy_types(execution_engine, expected_type):
    types = []
    type_module = _get_dialect_type_module(execution_engine=execution_engine)
    try:
        # bigquery geography requires installing an extra package
        if (
            expected_type.lower() == "geography"
            and execution_engine.engine.dialect.name.lower() == GXSqlDialect.BIGQUERY
            and not BIGQUERY_GEO_SUPPORT
        ):
            logger.warning(
                "BigQuery GEOGRAPHY type is not supported by default. "
                + "To install support, please run:"
                + "  $ pip install 'sqlalchemy-bigquery[geography]'"
            )
        elif type_module.__name__ == "pyathena.sqlalchemy_athena":
            potential_type = get_pyathena_potential_type(type_module, expected_type)
            # In the case of the PyAthena dialect we need to verify that
            # the type returned is indeed a type and not an instance.
            if not inspect.isclass(potential_type):
                real_type = type(potential_type)
            else:
                real_type = potential_type
            types.append(real_type)
        elif type_module.__name__ == "clickhouse_sqlalchemy.drivers.base":
            potential_type = get_clickhouse_sqlalchemy_potential_type(type_module, expected_type)
            types.append(potential_type)
        elif type_module.__name__ == "sqlalchemy_redshift.dialect":
            types.extend(_get_redshift_sqlalchemy_types(type_module, expected_type))
        else:
            potential_type = getattr(type_module, expected_type)
            types.append(potential_type)
    except AttributeError:
        logger.debug(f"Unrecognized type: {expected_type}")
    if len(types) == 0:
        logger.debug("No recognized sqlalchemy types in type_list for current dialect.")

    return types


def _get_redshift_sqlalchemy_types(
    type_module: ModuleType, expected_type: Any
) -> list[sa.sql.type_api.TypeEngine]:
    types: list[sa.sql.type_api.TypeEngine] = []
    potential_type = getattr(type_module, expected_type)
    types.append(potential_type)
    if expected_type.lower() == "decimal":
        # There is no redshift numeric type NUMERIC. It is suppose to be a synonym for
        # the official type DECIMAL, according to the docs:
        # https://docs.aws.amazon.com/redshift/latest/dg/c_Supported_data_types.html
        # However we have observed the raw sqltypes.[NUMERIC|Numeric] instead so we
        # add this as an allowed matching type.
        types.append(sa.sql.sqltypes.NUMERIC)
    return types


def _get_dialect_type_module(  # noqa: C901, PLR0911 # FIXME CoP
    execution_engine,
):
    if execution_engine.dialect_module is None:
        logger.warning("No sqlalchemy dialect found; relying in top-level sqlalchemy types.")
        return sa

    # Redshift does not (yet) export types to top level; only recognize base SA types
    if aws.redshiftdialect and isinstance(
        execution_engine.dialect_module,
        aws.redshiftdialect.RedshiftDialect,
    ):
        return execution_engine.dialect_module.sa
    else:
        pass

    # Bigquery works with newer versions, but use a patch if we had to define bigquery_types_tuple
    try:
        if BigQueryDialect and (
            isinstance(
                execution_engine.dialect_module,
                BigQueryDialect,
            )
            and bigquery_types_tuple is not None
        ):
            return bigquery_types_tuple
    except (TypeError, AttributeError):
        pass

    # Teradata types module
    try:
        if (
            issubclass(
                execution_engine.dialect_module,
                teradatasqlalchemy.dialect.TeradataDialect,
            )
            and teradatatypes is not None
        ):
            return teradatatypes
    except (TypeError, AttributeError):
        pass

    try:
        if (
            issubclass(
                execution_engine.dialect_module,
                clickhouse_sqlalchemy.drivers.base.ClickHouseDialect,
            )
            and ch_types is not None
        ):
            return ch_types
    except (TypeError, AttributeError):
        pass

    # Trino types module
    try:
        if (
            trino.trinodialect
            and trino.trinotypes
            and isinstance(
                execution_engine.dialect,
                trino.trinodialect.TrinoDialect,
            )
        ):
            return trino.trinotypes
    except (TypeError, AttributeError):
        pass

    return execution_engine.dialect_module


def _native_type_type_map(type_):  # noqa: C901, PLR0911 # FIXME CoP
    # We allow native python types in cases where the underlying type is "object":
    if type_.lower() == "none":
        return (type(None),)
    elif type_.lower() == "bool":
        return (bool,)
    elif type_.lower() in ["int", "long"]:
        return (int,)
    elif type_.lower() == "float":
        return (float,)
    elif type_.lower() == "bytes":
        return (bytes,)
    elif type_.lower() == "complex":
        return (complex,)
    elif type_.lower() in ["str", "string_types"]:
        return (str,)
    elif type_.lower() == "list":
        return (list,)
    elif type_.lower() == "dict":
        return (dict,)
    elif type_.lower() == "unicode":
        return None



import pytest

from great_expectations.data_context.store.json_site_store import JsonSiteStore


@pytest.mark.cloud
def test_gx_cloud_response_json_to_object_dict():
    response_json = {
        "data": {
            "id": "683df05f-efec-48d4-a45a-02e8318043b8",
            "attributes": {
                "rendered_data_doc": {
                    "sections": [],
                    "data_asset_name": "my_asset",
                }
            },
        }
    }

    actual = JsonSiteStore.gx_cloud_response_json_to_object_dict(response_json)
    expected = {
        "id": "683df05f-efec-48d4-a45a-02e8318043b8",
        "sections": [],
        "data_asset_name": "my_asset",
    }

    assert actual == expected



import re

import great_expectations as gx

context = gx.get_context()

datasource_name = "my_s3_datasource"
bucket_name = "great-expectations-docs-test"

datasource = context.data_sources.add_pandas_s3(
    name=datasource_name, bucket=bucket_name, boto3_options={}
)

asset_name = "my_taxi_data_asset"
s3_prefix = "data/taxi_yellow_tripdata_samples/"
data_asset = datasource.add_csv_asset(name=asset_name, s3_prefix=s3_prefix)

batching_regex = re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
batch_definition = data_asset.add_batch_definition_monthly("monthly", regex=batching_regex)

# Get all batches by month
batch_request = batch_definition.build_batch_request()
batch_identifiers_list = data_asset.get_batch_identifiers_list(batch_request)
assert len(batch_identifiers_list) == 3
assert batch_identifiers_list[0] == {
    "year": "2019",
    "month": "01",
    "path": "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01.csv",
}

# Get a specific batch by month
batch_request = batch_definition.build_batch_request({"year": "2019", "month": "02"})
batch = data_asset.get_batch(batch_request)
assert batch.metadata == {
    "year": "2019",
    "month": "02",
    "path": "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-02.csv",
}



from typing import List

from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

gcs_integration_tests = []

connecting_to_your_data: List[IntegrationTestFixture] = []

how_to_configure_metadata_store = [
    # Chetan - 20231117 - These have been commented out due to their reliance on the CLI (which has been deleted).  # noqa: E501 # FIXME CoP
    #                     They should be re-enabled once they have been updated.
    # IntegrationTestFixture(
    #     name="how_to_configure_an_expectation_store_in_gcs",
    #     user_flow_script="docs/docusaurus/docs/oss/guides/setup/configuring_metadata_stores/how_to_configure_an_expectation_store_in_gcs.py",  # noqa: E501 # FIXME CoP
    #     data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
    #     backend_dependencies=[BackendDependencies.GCS],
    # ),
    # IntegrationTestFixture(
    #     name="how_to_host_and_share_data_docs_on_gcs",
    #     user_flow_script="docs/docusaurus/docs/oss/guides/setup/configuring_data_docs/how_to_host_and_share_data_docs_on_gcs.py",  # noqa: E501 # FIXME CoP
    #     data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
    #     data_dir="tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
    #     backend_dependencies=[BackendDependencies.GCS],
    # ),
    # IntegrationTestFixture(
    #     name="how_to_configure_a_validation_result_store_in_gcs",
    #     user_flow_script="docs/docusaurus/docs/oss/guides/setup/configuring_metadata_stores/how_to_configure_a_validation_result_store_in_gcs.py",  # noqa: E501 # FIXME CoP
    #     data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
    #     data_dir="tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
    #     backend_dependencies=[BackendDependencies.GCS],
    # ),
]

partition_data = [
    IntegrationTestFixture(
        name="partition_data_on_datetime_gcs",
        user_flow_script="tests/integration/test_definitions/gcs/partitioned_on_datetime.py",
        data_context_dir=None,
        backend_dependencies=[BackendDependencies.GCS],
    ),
    IntegrationTestFixture(
        name="gcs_by_path",
        user_flow_script="tests/integration/test_definitions/gcs/select_batch_by_path.py",
        data_context_dir=None,
        backend_dependencies=[BackendDependencies.GCS],
    ),
]

sample_data: List[IntegrationTestFixture] = []

deployment_patterns = [
    IntegrationTestFixture(
        name="deployment_patterns_file_gcs",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/database/gcp_deployment_patterns_file_gcs.py",
        data_context_dir=None,
        backend_dependencies=[BackendDependencies.GCS],
    ),
]

fluent_datasources = [
    IntegrationTestFixture(
        name="how_to_connect_to_data_on_gcs_using_pandas",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/filesystem/how_to_connect_to_data_on_gcs_using_pandas.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.GCS],
    ),
]

gcs_integration_tests += connecting_to_your_data
gcs_integration_tests += how_to_configure_metadata_store
gcs_integration_tests += partition_data
gcs_integration_tests += sample_data
gcs_integration_tests += deployment_patterns
gcs_integration_tests += fluent_datasources



from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING, Literal, Mapping, Optional, Union

from ruamel.yaml import YAML, YAMLError
from ruamel.yaml.constructor import DuplicateKeyError

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.data_context.serializable_data_context import (
    SerializableDataContext,
)
from great_expectations.data_context.data_context_variables import (
    DataContextVariableSchema,
    FileDataContextVariables,
)
from great_expectations.data_context.types.base import (
    DataContextConfig,
)
from great_expectations.datasource.fluent.config import GxConfig

if TYPE_CHECKING:
    from great_expectations.alias_types import JSONValues, PathStr
    from great_expectations.core.config_provider import _ConfigurationProvider
    from great_expectations.data_context.store.datasource_store import DatasourceStore

logger = logging.getLogger(__name__)
yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False


@public_api
class FileDataContext(SerializableDataContext):
    """Subclass of AbstractDataContext that contains functionality necessary to work in a filesystem-backed environment."""  # noqa: E501 # FIXME CoP

    def __init__(
        self,
        project_config: Optional[DataContextConfig] = None,
        context_root_dir: Optional[PathStr] = None,
        project_root_dir: Optional[PathStr] = None,
        runtime_environment: Optional[dict] = None,
        user_agent_str: Optional[str] = None,
    ) -> None:
        """FileDataContext constructor

        Args:
            project_config (DataContextConfig):  Config for current DataContext
            context_root_dir (Optional[str]): location to look for the ``great_expectations.yml`` file. If None,
                searches for the file based on conventions for project subdirectories.
            runtime_environment (Optional[dict]): a dictionary of config variables that override both those set in
                config_variables.yml and the environment
        """  # noqa: E501 # FIXME CoP
        self._context_root_directory = self._init_context_root_directory(
            context_root_dir=context_root_dir,
            project_root_dir=project_root_dir,
        )
        self._scaffold_project()

        self._project_config = self._init_project_config(project_config)
        super().__init__(
            context_root_dir=self._context_root_directory,
            runtime_environment=runtime_environment,
            user_agent_str=user_agent_str,
        )

    @property
    @override
    def mode(self) -> Literal["file"]:
        return "file"

    def _init_context_root_directory(
        self, context_root_dir: Optional[PathStr], project_root_dir: Optional[PathStr]
    ) -> str:
        context_root_dir = self._resolve_context_root_dir_and_project_root_dir(
            context_root_dir=context_root_dir, project_root_dir=project_root_dir
        )

        if isinstance(context_root_dir, pathlib.Path):
            context_root_dir = str(context_root_dir)

        if not context_root_dir:
            context_root_dir = self.find_context_root_dir()

        return context_root_dir

    def _scaffold_project(self) -> None:
        """Prepare a `great_expectations` directory with all necessary subdirectories.
        If one already exists, no-op.
        """
        if self.is_project_scaffolded(self._context_root_directory):
            return

        # GX makes an important distinction between project directory and context directory.
        # The former corresponds to the root of the user's project while the latter
        # encapsulates any config (in the form of a great_expectations/ directory).
        project_root_dir = pathlib.Path(self._context_root_directory).parent
        relative_context_dir = pathlib.Path(self._context_root_directory).name
        self._scaffold(
            project_root_dir=project_root_dir,
            context_root_dir_name=relative_context_dir,
        )

    @override
    def _init_project_config(
        self, project_config: Optional[Union[DataContextConfig, Mapping]]
    ) -> DataContextConfig:
        if project_config:
            project_config = FileDataContext.get_or_create_data_context_config(project_config)
        else:
            project_config = FileDataContext._load_file_backed_project_config(
                context_root_directory=self._context_root_directory,
            )
        return project_config

    @override
    def _init_datasource_store(self) -> DatasourceStore:
        from great_expectations.data_context.store.datasource_store import (
            DatasourceStore,
        )

        store_name: str = "datasource_store"  # Never explicitly referenced but adheres
        # to the convention set by other internal Stores
        store_backend: dict = {
            "class_name": "InlineStoreBackend",
            "resource_type": DataContextVariableSchema.DATASOURCES,
        }
        runtime_environment: dict = {
            "root_directory": self.root_directory,
            "data_context": self,
            # By passing this value in our runtime_environment,
            # we ensure that the same exact context (memory address and all) is supplied to the Store backend  # noqa: E501 # FIXME CoP
        }

        datasource_store = DatasourceStore(
            store_name=store_name,
            store_backend=store_backend,
            runtime_environment=runtime_environment,
        )
        return datasource_store

    @override
    def _init_variables(self) -> FileDataContextVariables:
        variables = FileDataContextVariables(
            config=self._project_config,
            config_provider=self.config_provider,
            data_context=self,
        )
        return variables

    @override
    def _save_project_config(self) -> None:
        """
        See parent 'AbstractDataContext._save_project_config()` for more information.

        Explicitly override base class implementation to retain legacy behavior.
        """
        config_filepath = pathlib.Path(self.root_directory, self.GX_YML)

        logger.debug(
            f"Starting DataContext._save_project_config; attempting to update {config_filepath}"
        )

        try:
            with open(config_filepath, "w") as outfile:
                fluent_datasources = self._synchronize_fluent_datasources()
                if fluent_datasources:
                    self.fluent_config.update_datasources(datasources=fluent_datasources)
                    logger.info(
                        f"Saving {len(self.fluent_config.datasources)} Fluent Datasources to {config_filepath}"  # noqa: E501 # FIXME CoP
                    )
                    fluent_json_dict: dict[str, JSONValues] = self.fluent_config._json_dict()
                    fluent_json_dict = (
                        self.fluent_config._exclude_name_fields_from_fluent_datasources(
                            config=fluent_json_dict
                        )
                    )
                    self.config._commented_map.update(fluent_json_dict)

                self.config.to_yaml(outfile)
        except PermissionError as e:
            logger.warning(f"Could not save project config to disk: {e}")

    @classmethod
    def _load_file_backed_project_config(
        cls,
        context_root_directory: PathStr,
    ) -> DataContextConfig:
        path_to_yml = pathlib.Path(context_root_directory, cls.GX_YML)
        try:
            with open(path_to_yml) as data:
                config_commented_map_from_yaml = yaml.load(data)

        except DuplicateKeyError:
            raise gx_exceptions.InvalidConfigurationYamlError(  # noqa: TRY003 # FIXME CoP
                "Error: duplicate key found in project YAML file."
            )
        except YAMLError as err:
            raise gx_exceptions.InvalidConfigurationYamlError(  # noqa: TRY003 # FIXME CoP
                f"Your configuration file is not a valid yml file likely due to a yml syntax error:\n\n{err}"  # noqa: E501 # FIXME CoP
            )
        except OSError:
            raise gx_exceptions.ConfigNotFoundError()

        try:
            return DataContextConfig.from_commented_map(
                commented_map=config_commented_map_from_yaml
            )
        except gx_exceptions.InvalidDataContextConfigError:  # noqa: TRY203 # FIXME CoP
            # Just to be explicit about what we intended to catch
            raise

    @override
    def _load_fluent_config(self, config_provider: _ConfigurationProvider) -> GxConfig:
        logger.info(f"{type(self).__name__} loading fluent config")
        if not self.root_directory:
            logger.warning("`root_directory` not set, cannot load fluent config")
        else:
            path_to_fluent_yaml = pathlib.Path(self.root_directory) / self.GX_YML
            if path_to_fluent_yaml.exists():
                gx_config = GxConfig.parse_yaml(path_to_fluent_yaml, _allow_empty=True)

                for datasource in gx_config.datasources:
                    datasource._data_context = self

                return gx_config
            logger.info(f"no fluent config at {path_to_fluent_yaml.absolute()}")
        return GxConfig(fluent_datasources=[])



from __future__ import annotations

import datetime
import json
import re
from collections import OrderedDict
from string import Template as pTemplate
from typing import TYPE_CHECKING, Any, ClassVar, Iterable, Mapping
from uuid import uuid4

import mistune
from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
    select_autoescape,
)

from great_expectations.compatibility.typing_extensions import override

try:
    from jinja2 import contextfilter  # type: ignore[attr-defined] # for jinja 2.0
except ImportError:
    from jinja2 import pass_context as contextfilter

from great_expectations import __version__ as ge_version
from great_expectations.render import (
    RenderedComponentContent,
    RenderedContent,
    RenderedDocumentContent,
)

if TYPE_CHECKING:
    from jinja2 import BaseLoader
    from jinja2 import Template as jTemplate


class PrettyPrintTemplate:
    def render(self, document, indent=2) -> None:
        print(json.dumps(document, indent=indent))


# Abe 2019/06/26: This View should probably actually be called JinjaView or something similar.
# Down the road, I expect to wind up with class hierarchy along the lines of:
#   View > JinjaView > GEContentBlockJinjaView
class DefaultJinjaView:
    """
    Defines a method for converting a document to human-consumable form

    Dependencies
    ~~~~~~~~~~~~
    * Font Awesome 5.10.1
    * Bootstrap 4.3.1
    * jQuery 3.2.1
    * Vega 5
    * Vega-Lite 4
    * Vega-Embed 6

    """

    _template: ClassVar[str]

    def __init__(self, custom_styles_directory=None, custom_views_directory=None) -> None:
        self.custom_styles_directory = custom_styles_directory
        self.custom_views_directory = custom_views_directory

        templates_loader = PackageLoader("great_expectations", "render/view/templates")
        styles_loader = PackageLoader("great_expectations", "render/view/static/styles")

        loaders: list[BaseLoader] = [templates_loader, styles_loader]
        if self.custom_styles_directory:
            loaders.append(FileSystemLoader(self.custom_styles_directory))
        if self.custom_views_directory:
            loaders.append(FileSystemLoader(self.custom_views_directory))

        self.env = Environment(
            loader=ChoiceLoader(loaders),
            autoescape=select_autoescape(["html", "xml"]),
            extensions=["jinja2.ext.do"],
        )

        self.env.filters["render_string_template"] = self.render_string_template
        self.env.filters["render_styling_from_string_template"] = (
            self.render_styling_from_string_template
        )
        self.env.filters["render_styling"] = self.render_styling
        self.env.filters["render_content_block"] = self.render_content_block
        self.env.filters["render_markdown"] = self.render_markdown
        self.env.filters["get_html_escaped_json_string_from_dict"] = (
            self.get_html_escaped_json_string_from_dict
        )
        self.env.filters["generate_html_element_uuid"] = self.generate_html_element_uuid
        self.env.filters["attributes_dict_to_html_string"] = self.attributes_dict_to_html_string
        self.env.filters["render_bootstrap_table_data"] = self.render_bootstrap_table_data
        self.env.globals["ge_version"] = ge_version
        self.env.filters["add_data_context_id_to_url"] = self.add_data_context_id_to_url

    def render(self, document, template=None, **kwargs):
        self._validate_document(document)

        if template is None:
            template = self._template

        t = self._get_template(template)
        if isinstance(document, RenderedContent):
            document = document.to_json_dict()
        return t.render(document, **kwargs)

    def _get_template(self, template_str: str) -> jTemplate:
        template = self.env.get_template(template_str)
        template.globals["now"] = lambda: datetime.datetime.now(datetime.timezone.utc)

        return template

    @contextfilter  # type: ignore[misc] # untyped 3rd party decorator
    def add_data_context_id_to_url(
        self, jinja_context: Any, url: str, add_datetime: bool = True
    ) -> str:
        data_context_id = jinja_context.get("data_context_id")
        if add_datetime:
            datetime_iso_string = datetime.datetime.now(datetime.timezone.utc).strftime(
                "%Y%m%dT%H%M%S.%fZ"
            )
            url += f"?d={datetime_iso_string}"
        if data_context_id:
            url = f"{url}&dataContextId=" if add_datetime else f"{url}?dataContextId="
            url += str(data_context_id)
        return url

    @contextfilter  # type: ignore[misc] # untyped 3rd party decorator
    def render_content_block(  # noqa: C901, PLR0911, PLR0912 # FIXME CoP
        self,
        jinja_context: Any,
        content_block: str | list | dict | RenderedComponentContent,
        index: Any = None,
        content_block_id: Any = None,
        render_to_markdown: bool = False,
    ) -> RenderedComponentContent | dict | str:
        """

        :param jinja_context:
        :param content_block:
        :param index:
        :param content_block_id:
        :param render_to_markdown: Whether this method should render the markdown version instead of HTML
        :return:
        """  # noqa: E501 # FIXME CoP
        if isinstance(content_block, str):
            return content_block
        elif content_block is None:
            return ""
        elif isinstance(content_block, list):
            # If the content_block item here is actually a list of content blocks then we want to recursively render  # noqa: E501 # FIXME CoP
            rendered_block = ""
            for idx, content_block_el in enumerate(content_block):
                if isinstance(content_block_el, RenderedComponentContent) or (
                    isinstance(content_block_el, dict) and "content_block_type" in content_block_el
                ):
                    new_content_block_id = None
                    if content_block_id:
                        new_content_block_id = f"{content_block_id}-{idx!s}"
                    rendered_block += self.render_content_block(
                        jinja_context,
                        content_block_el,
                        idx,
                        content_block_id=new_content_block_id,
                    )
                else:  # noqa: PLR5501 # FIXME CoP
                    if render_to_markdown:
                        rendered_block += str(content_block_el)
                    else:
                        rendered_block += f"<span>{content_block_el!s}</span>"
            return rendered_block
        elif not isinstance(content_block, dict):
            return content_block
        content_block_type = content_block.get("content_block_type")
        if content_block_type is None:
            return content_block

        if render_to_markdown:
            template_filename = f"markdown_{content_block_type}.j2"
        else:
            template_filename = f"{content_block_type}.j2"
        template = self._get_template(template_str=template_filename)
        if content_block_id:
            return template.render(
                jinja_context,
                content_block=content_block,
                index=index,
                content_block_id=content_block_id,
            )
        else:
            return template.render(jinja_context, content_block=content_block, index=index)

    def render_dict_values(
        self, context: Any, dict_: dict, index: Any = None, content_block_id: Any = None
    ) -> None:
        for key, val in dict_.items():
            if key.startswith("_"):
                continue
            dict_[key] = self.render_content_block(context, val, index, content_block_id)

    @contextfilter  # type: ignore[misc] # untyped 3rd party decorator
    def render_bootstrap_table_data(
        self,
        context: Any,
        table_data: Iterable[dict],
        index: Any | None = None,
        content_block_id: str | None = None,
    ):
        for table_data_dict in table_data:
            self.render_dict_values(context, table_data_dict, index, content_block_id)
        return table_data

    def get_html_escaped_json_string_from_dict(self, source_dict: dict) -> str:
        return json.dumps(source_dict).replace('"', '\\"').replace('"', "&quot;")

    def attributes_dict_to_html_string(self, attributes_dict: dict, prefix=""):
        attributes_string = ""
        if prefix:
            prefix += "-"
        for attribute, value in attributes_dict.items():
            attributes_string += f'{prefix}{attribute}="{value}" '
        return attributes_string

    def render_styling(self, styling: Mapping) -> str:
        """Adds styling information suitable for an html tag.

        Example styling block::

            styling = {
                "classes": ["alert", "alert-warning"],
                "attributes": {
                    "role": "alert",
                    "data-toggle": "popover",
                },
                "styles" : {
                    "padding" : "10px",
                    "border-radius" : "2px",
                }
            }

        The above block returns a string similar to::

            'class="alert alert-warning" role="alert" data-toggle="popover" style="padding: 10px; border-radius: 2px"'

        "classes", "attributes" and "styles" are all optional parameters.
        If they aren't present, they simply won't be rendered.

        Other dictionary keys are also allowed and ignored.
        """  # noqa: E501 # FIXME CoP

        class_list = styling.get("classes", None)
        if class_list is None:
            class_str = ""
        else:
            if type(class_list) == str:  # noqa: E721 # FIXME CoP
                raise TypeError("classes must be a list, not a string.")  # noqa: TRY003 # FIXME CoP
            class_str = f'class="{" ".join(class_list)}" '

        attribute_dict = styling.get("attributes", None)
        if attribute_dict is None:
            attribute_str = ""
        else:
            attribute_str = ""
            for k, v in attribute_dict.items():
                attribute_str += f'{k}="{v}" '

        style_dict = styling.get("styles", None)
        if style_dict is None:
            style_str = ""
        else:
            style_str = 'style="'
            style_str += " ".join([f"{k}:{v};" for k, v in style_dict.items()])
            style_str += '" '

        styling_string = pTemplate("$classes$attributes$style").substitute(
            {
                "classes": class_str,
                "attributes": attribute_str,
                "style": style_str,
            }
        )

        return styling_string

    def render_styling_from_string_template(self, template: dict | OrderedDict) -> str:
        # NOTE: We should add some kind of type-checking to template
        """This method is a thin wrapper use to call `render_styling` from within jinja templates."""  # noqa: E501 # FIXME CoP
        if not isinstance(template, (dict, OrderedDict)):
            return template

        if "styling" in template:
            return self.render_styling(template["styling"])

        else:
            return ""

    def generate_html_element_uuid(self, prefix=None):
        if prefix:
            return prefix + str(uuid4())
        else:
            return str(uuid4())

    def render_markdown(self, markdown):
        try:
            return mistune.markdown(markdown)
        except OSError:
            return markdown

    def render_string_template(self, template):  # noqa: C901, PLR0912 # FIXME CoP
        # NOTE: Using this line for debugging. This should probably be logged...?
        # print(template)

        # NOTE: We should add some kind of type-checking to template
        if not isinstance(template, (dict, OrderedDict)):
            return template

        # if there are any groupings of two or more $, we need to double the groupings to account
        # for template string substitution escaping
        template["template"] = re.sub(
            r"\${2,}", lambda m: m.group(0) * 2, template.get("template", "")
        )

        tag = template.get("tag", "span")
        template["template"] = template.get("template", "").replace("$PARAMETER", "$$PARAMETER")
        template["template"] = template.get("template", "").replace("\n", "<br>")

        if "tooltip" in template:
            if template.get("styling", {}).get("classes"):
                classes = template.get("styling", {}).get("classes")
                classes.append("cooltip")
                template["styling"]["classes"] = classes
            elif template.get("styling"):
                template["styling"]["classes"] = ["cooltip"]
            else:
                template["styling"] = {"classes": ["cooltip"]}

            tooltip_content = template["tooltip"]["content"]
            tooltip_content.replace("\n", "<br>")
            placement = template["tooltip"].get("placement", "top")
            base_template_string = f"""
                <{tag} $styling>
                    $template
                    <span class={placement}>
                        {tooltip_content}
                    </span>
                </{tag}>
            """
        else:
            base_template_string = f"""
                <{tag} $styling>
                    $template
                </{tag}>
            """

        if "styling" in template:
            params = template.get("params", {})

            # Apply default styling
            if "default" in template["styling"]:
                default_parameter_styling = template["styling"]["default"]
                default_param_tag = default_parameter_styling.get("tag", "span")
                base_param_template_string = (
                    f"<{default_param_tag} $styling>$content</{default_param_tag}>"
                )

                for parameter in template["params"]:
                    # If this param has styling that over-rides the default, skip it here and get it in the next loop.  # noqa: E501 # FIXME CoP
                    if "params" in template["styling"]:
                        if parameter in template["styling"]["params"]:
                            continue

                    params[parameter] = pTemplate(base_param_template_string).safe_substitute(
                        {
                            "styling": self.render_styling(default_parameter_styling),
                            "content": params[parameter],
                        }
                    )

            # Apply param-specific styling
            if "params" in template["styling"]:
                # params = template["params"]
                for parameter, parameter_styling in template["styling"]["params"].items():
                    if parameter not in params:
                        continue
                    param_tag = parameter_styling.get("tag", "span")
                    param_template_string = f"<{param_tag} $styling>$content</{param_tag}>"
                    params[parameter] = pTemplate(param_template_string).safe_substitute(
                        {
                            "styling": self.render_styling(parameter_styling),
                            "content": params[parameter],
                        }
                    )

            string = pTemplate(
                pTemplate(base_template_string).safe_substitute(
                    {
                        "template": template["template"],
                        "styling": self.render_styling(template.get("styling", {})),
                    }
                )
            ).safe_substitute(params)
            return string

        return pTemplate(
            pTemplate(base_template_string).safe_substitute(
                {
                    "template": template.get("template", ""),
                    "styling": self.render_styling(template.get("styling", {})),
                }
            )
        ).safe_substitute(template.get("params", {}))

    def _validate_document(self, document) -> None:
        raise NotImplementedError


class DefaultJinjaPageView(DefaultJinjaView):
    _template = "page.j2"

    @override
    def _validate_document(self, document: RenderedDocumentContent) -> None:
        assert isinstance(document, RenderedDocumentContent)


class DefaultJinjaIndexPageView(DefaultJinjaPageView):
    _template = "index_page.j2"


class DefaultJinjaSectionView(DefaultJinjaView):
    _template = "section.j2"

    @override
    def _validate_document(self, document: Mapping) -> None:
        assert isinstance(document["section"], dict)  # For now low-level views take dicts


class DefaultJinjaComponentView(DefaultJinjaView):
    _template = "component.j2"

    @override
    def _validate_document(self, document: Mapping) -> None:
        assert isinstance(document["content_block"], dict)  # For now low-level views take dicts


class DefaultMarkdownPageView(DefaultJinjaView):
    """
    Convert a document to markdown format.
    """

    @override
    def _validate_document(self, document: RenderedDocumentContent) -> None:
        """
        Validate that the document is of the appropriate type at runtime.
        """
        assert isinstance(document, RenderedDocumentContent)

    _template = "markdown_validation_results_page.j2"

    @override
    def render(self, document: list, template=None, **kwargs) -> list[str] | str:
        """
        Handle list as well as single document
        """
        if isinstance(document, list):
            # We need to keep this as super(DefaultMarkdownPageView, self); otherwise a wrong render will be called.  # noqa: E501 # FIXME CoP
            return [
                super(DefaultMarkdownPageView, self).render(document=d, template=template, **kwargs)
                for d in document
            ]

        else:
            return super().render(document=document, template=template, **kwargs)

    @override
    def render_string_template(self, template: pTemplate | dict | OrderedDict) -> pTemplate | str:
        """
        Render string for markdown rendering. Bold all parameters and perform substitution.
        Args:
            template: python Template object

        Returns:
            Template with substituted values and all parameters bolded

        """

        if not isinstance(template, (dict, OrderedDict)):
            return template

        # replace and render any horizontal lines using ***
        tag = template.get("tag", None)
        if tag and tag == "hr":
            template["template"] = "***"

        # if there are any groupings of two or more $, we need to double the groupings to account
        # for template string substitution escaping
        template["template"] = re.sub(
            r"\${2,}", lambda m: m.group(0) * 2, template.get("template", "")
        )

        # Bold all parameters:
        base_param_template_string = "**$content**"

        # Make sure template["params"] is a dict
        template["params"] = template.get("params", {})

        # TODO: Revisit handling of icons in markdown. E.g. inline rendered icons.
        if "markdown_status_icon" in template["params"]:
            return template["params"]["markdown_status_icon"]

        for parameter in template["params"]:
            if parameter == "html_success_icon":
                template["params"][parameter] = ""
                continue
            # to escape any values that are '*' which, when combined with bold ('**') in markdown,
            # does not give the output we want.
            elif template["params"][parameter] == "*":
                template["params"][parameter] = "\\*"
                continue

            template["params"][parameter] = pTemplate(base_param_template_string).safe_substitute(
                {
                    "content": template["params"][parameter],
                }
            )

        template["template"] = template.get("template", "").replace("$PARAMETER", "$$PARAMETER")

        return pTemplate(template["template"]).safe_substitute(template.get("params", {}))

    @override
    @contextfilter  # type: ignore[misc] # untyped 3rd party decorator
    def render_content_block(
        self,
        jinja_context: Any,
        content_block: Any,
        index: Any | None = None,
        content_block_id: str | None = None,
        render_to_markdown: bool = True,
    ) -> RenderedComponentContent | dict | str:
        """
        Render a content block to markdown using jinja templates.
        Args:
            jinja_context:
            content_block:
            index:
            content_block_id:
            render_to_markdown: Default of True here instead of parent class default of False

        Returns:

        """

        return super().render_content_block(
            jinja_context=jinja_context,
            content_block=content_block,
            index=index,
            content_block_id=content_block_id,
            render_to_markdown=render_to_markdown,
        )



"""
POC for dynamically bootstrapping context.sources with Datasource factory methods.
"""

from __future__ import annotations

import logging
from pprint import pformat as pf
from typing import Set, Type

from great_expectations.compatibility.pydantic import ModelMetaclass
from great_expectations.datasource.fluent.sources import DataSourceManager
from great_expectations.datasource.fluent.type_lookup import TypeLookup

logger = logging.getLogger(__name__)


class MetaDatasource(ModelMetaclass):
    __cls_set: Set[Type] = set()

    def __new__(  # noqa: PYI034 # Self cannot be used with Metaclass
        meta_cls: Type[MetaDatasource], cls_name: str, bases: tuple[type], cls_dict
    ) -> MetaDatasource:
        """
        MetaDatasource hook that runs when a new `Datasource` is defined.
        This methods binds a factory method for the defined `Datasource` to `DataSourceManager` class which becomes
        available as part of the `DataContext`.

        Also binds asset adding methods according to the declared `asset_types`.
        """  # noqa: E501 # FIXME CoP
        logger.debug(f"1a. {meta_cls.__name__}.__new__() for `{cls_name}`")

        cls = super().__new__(meta_cls, cls_name, bases, cls_dict)

        if cls_name in ("Datasource", "InvalidDatasource") or cls_name.startswith("_"):
            # NOTE: the above check is brittle and must be kept in-line with the Datasource.__name__
            logger.debug(f"1c. Skip factory registration of base `{cls_name}`")
            return cls

        logger.debug(f"  {cls_name} __dict__ ->\n{pf(cls.__dict__, depth=3)}")

        meta_cls.__cls_set.add(cls)
        logger.debug(f"Datasources: {len(meta_cls.__cls_set)}")

        if cls.__module__ == "__main__":
            logger.warning(
                f"Datasource `{cls_name}` should not be defined as part of __main__ this may cause typing lookup collisions"  # noqa: E501 # FIXME CoP
            )
        # instantiate new TypeLookup to prevent child classes conflicts with parent class asset types  # noqa: E501 # FIXME CoP
        cls._type_lookup = TypeLookup()
        DataSourceManager.register_datasource(cls)
        return cls



"""
https://learn.microsoft.com/en-us/python/api/semantic-link-sempy/sempy.fabric?view=semantic-link-python
"""

from __future__ import annotations

import logging
import os
import uuid
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Dict,
    Final,
    List,
    Literal,
    Optional,
    Set,
    Type,
    Union,
)

from typing_extensions import Annotated, TypeAlias

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.batch_spec import FabricBatchSpec
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.batch_identifier_util import make_batch_identifier
from great_expectations.datasource.fluent.constants import _DATA_CONNECTOR_NAME
from great_expectations.datasource.fluent.interfaces import (
    Batch,
    DataAsset,
    Datasource,
    Sorter,
    TestConnectionError,
)
from great_expectations.exceptions.exceptions import BuildBatchRequestError

if TYPE_CHECKING:
    from great_expectations.core.batch_spec import FabricReaderMethods
    from great_expectations.core.partitioners import ColumnPartitioner
    from great_expectations.datasource.fluent import BatchParameters
    from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice
    from great_expectations.datasource.fluent.interfaces import (
        BatchMetadata,
    )
    from great_expectations.execution_engine import PandasExecutionEngine

LOGGER = logging.getLogger(__name__)

SortersDefinition: TypeAlias = List[Union[Sorter, str, dict]]

_REQUIRED_FABRIC_SERVICE: Final[str] = "Microsoft.ProjectArcadia"
Mode: TypeAlias = Literal["xmla", "rest", "onelake"]


class _PowerBIAsset(DataAsset):
    """Microsoft PowerBI Asset base class."""

    _reader_method: ClassVar[FabricReaderMethods]
    _EXCLUDE_FROM_READER_OPTIONS: ClassVar[Set[str]] = {
        "batch_definitions",
        "batch_metadata",
        "name",
        "order_by",
        "type",
        "id",
    }

    @override
    def test_connection(self) -> None:
        """
        Whatever is needed to test the connection to and/or validity of the asset.
        This could be a noop.
        """
        LOGGER.debug(f"Testing connection to {self.__class__.__name__} has not been implemented")

    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        return [IDDict(batch_request.options)]

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        self._validate_batch_request(batch_request)

        reader_options = {
            "workspace": self._datasource.workspace,
            "dataset": self._datasource.dataset,
            **self.dict(
                exclude=self._EXCLUDE_FROM_READER_OPTIONS,
                exclude_none=True,
                exclude_unset=True,
                by_alias=True,
                config_provider=self._datasource._config_provider,
            ),
        }

        batch_spec = FabricBatchSpec(
            reader_method=self._reader_method, reader_options=reader_options
        )
        # TODO: update get_batch_data_and_markers types
        execution_engine: PandasExecutionEngine = self.datasource.get_execution_engine()
        data, markers = execution_engine.get_batch_data_and_markers(batch_spec=batch_spec)

        # batch_definition (along with batch_spec and markers) is only here to satisfy a
        # legacy constraint when computing usage statistics in a validator. We hope to remove
        # it in the future.
        batch_definition = LegacyBatchDefinition(
            datasource_name=self.datasource.name,
            data_connector_name=_DATA_CONNECTOR_NAME,
            data_asset_name=self.name,
            batch_identifiers=make_batch_identifier(batch_request.options),
            batch_spec_passthrough=None,
        )

        batch_metadata: BatchMetadata = self._get_batch_metadata_from_batch_request(
            batch_request=batch_request, ignore_options=("dataframe",)
        )

        return Batch(
            datasource=self.datasource,
            data_asset=self,
            batch_request=batch_request,
            data=data,
            metadata=batch_metadata,
            batch_markers=markers,
            batch_spec=batch_spec.to_json_dict(),  # type: ignore[arg-type] # will be coerced to BatchSpec
            batch_definition=batch_definition,
        )

    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> BatchRequest:
        """A batch request that can be used to obtain batches for this DataAsset.

        Args:
            options: This is not currently supported and must be {} or None for this data asset.
            batch_slice: This is not currently supported and must be None for this data asset.
            partitioner: This is not currently supported and must be None for this data asset.

        Returns:
            A BatchRequest object that can be used to obtain a batch from an Asset by calling the
            get_batch method.
        """
        asset_type_name: str = self.__class__.__name__
        if options:
            raise BuildBatchRequestError(
                message=f"options is not currently supported for {asset_type_name} "
                "and must be None or {}."
            )

        if batch_slice is not None:
            raise BuildBatchRequestError(
                message=f"batch_slice is not currently supported for {asset_type_name} "
                "and must be None."
            )

        if partitioner is not None:
            raise BuildBatchRequestError(
                message=f"partitioner is not currently supported for {asset_type_name} "
                "and must be None."
            )

        return BatchRequest(
            datasource_name=self.datasource.name,
            data_asset_name=self.name,
            options={},
        )

    @override
    def _validate_batch_request(self, batch_request: BatchRequest) -> None:
        """Validates the batch_request has the correct form.

        Args:
            batch_request: A batch request object to be validated.
        """
        if not (
            batch_request.datasource_name == self.datasource.name
            and batch_request.data_asset_name == self.name
            and not batch_request.options
        ):
            expect_batch_request_form = BatchRequest[None](
                datasource_name=self.datasource.name,
                data_asset_name=self.name,
                options={},
                batch_slice=batch_request._batch_slice_input,  # type: ignore[attr-defined] # private attr does exist
            )
            raise gx_exceptions.InvalidBatchRequestError(  # noqa: TRY003 # FIXME CoP
                "BatchRequest should have form:\n"
                f"{pf(expect_batch_request_form.dict())}\n"
                f"but actually has form:\n{pf(batch_request.dict())}\n"
            )


@public_api
class PowerBIDax(_PowerBIAsset):
    """Microsoft PowerBI DAX."""

    _reader_method: ClassVar[FabricReaderMethods] = "evaluate_dax"

    type: Literal["powerbi_dax"] = "powerbi_dax"
    dax_string: str


@public_api
class PowerBIMeasure(_PowerBIAsset):
    """Microsoft PowerBI Measure."""

    _reader_method: ClassVar[FabricReaderMethods] = "evaluate_measure"

    type: Literal["powerbi_measure"] = "powerbi_measure"
    measure: Union[str, List[str]]
    groupby_columns: Optional[List[str]] = None
    filters: Optional[Dict[str, List[str]]] = None
    fully_qualified_columns: Optional[bool] = None
    num_rows: Optional[int] = None
    use_xmla: bool = False


@public_api
class PowerBITable(_PowerBIAsset):
    """Microsoft PowerBI Table."""

    _reader_method: ClassVar[FabricReaderMethods] = "read_table"

    type: Literal["powerbi_table"] = "powerbi_table"
    table: str
    fully_qualified_columns: bool = False
    num_rows: Optional[int] = None
    multiindex_hierarchies: bool = False
    mode: Mode = "xmla"


# This improves our error messages by providing a more specific type for pydantic to validate against  # noqa: E501 # FIXME CoP
# It also ensure the generated jsonschema has a oneOf instead of anyOf field for assets
# https://docs.pydantic.dev/1.10/usage/types/#discriminated-unions-aka-tagged-unions
AssetTypes = Annotated[
    Union[PowerBITable, PowerBIMeasure, PowerBIDax],
    pydantic.Field(discriminator="type"),
]


@public_api
class FabricPowerBIDatasource(Datasource):
    """
    Microsoft Fabric Datasource.

    https://pypi.org/project/semantic-link/
    """

    # class var definitions
    asset_types: ClassVar[List[Type[DataAsset]]] = [
        PowerBIDax,
        PowerBIMeasure,
        PowerBITable,
    ]
    # any fabric datsource specific fields should be added to this set
    # example a connection_string field or a data directory field
    _EXTRA_EXCLUDED_EXEC_ENG_ARGS: ClassVar[set] = {"workspace", "dataset"}

    # right side of the operator determines the type name
    # left side enforces the names on instance creation
    type: Literal["fabric_powerbi"] = "fabric_powerbi"
    assets: List[AssetTypes] = []

    # fabric datasource specific fields
    workspace: Optional[Union[uuid.UUID, str]] = None
    dataset: Union[uuid.UUID, str]

    @property
    @override
    def execution_engine_type(self) -> Type[PandasExecutionEngine]:
        """Return the PandasExecutionEngine unless the override is set"""
        from great_expectations.execution_engine.pandas_execution_engine import (
            PandasExecutionEngine,
        )

        return PandasExecutionEngine

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the FabricPowerBIDatasource.

        Args:
            test_assets: If assets have been passed to the Datasource, whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        if not self._running_on_fabric():
            raise TestConnectionError("Must be running Microsoft Fabric to use this datasource")  # noqa: TRY003 # FIXME CoP

        try:
            from sempy import fabric  # noqa: F401 # test if fabric is installed
        except Exception as import_err:
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                "Could not import `sempy.fabric`\npip install semantic-link-sempy"
            ) from import_err

        if self.assets and test_assets:
            for asset in self.assets:
                asset._datasource = self
                asset.test_connection()

    @public_api
    def add_powerbi_dax_asset(
        self,
        name: str,
        dax_string: str,
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> PowerBIDax:
        """Adds a PowerBIDax asset to this datasource.

        Args:
            name: The name of this asset.
            TODO: other args
            batch_metadata: BatchMetadata we want to associate with this DataAsset and all batches derived from it.

        Returns:
            The asset that is added to the datasource.
        """  # noqa: E501 # FIXME CoP
        asset = PowerBIDax(
            name=name,
            batch_metadata=batch_metadata or {},
            dax_string=dax_string,
        )
        return self._add_asset(asset)

    @public_api
    def add_powerbi_measure_asset(  # noqa: PLR0913 # FIXME CoP
        self,
        name: str,
        measure: Union[str, List[str]],
        batch_metadata: Optional[BatchMetadata] = None,
        groupby_columns: Optional[List[str]] = None,
        filters: Optional[Dict[str, List[str]]] = None,
        fully_qualified_columns: Optional[bool] = None,
        num_rows: Optional[int] = None,
        use_xmla: bool = False,
    ) -> PowerBIMeasure:
        """Adds a PowerBIMeasure asset to this datasource.

        Args:
            name: The name of this asset.
            batch_metadata: BatchMetadata we want to associate with this DataAsset and all batches derived from it.

        Returns:
            The asset that is added to the datasource.
        """  # noqa: E501 # FIXME CoP
        asset = PowerBIMeasure(
            name=name,
            batch_metadata=batch_metadata or {},
            groupby_columns=groupby_columns,
            measure=measure,
            # TODO: require custom serde for keys that are tuples
            filters=filters,
            fully_qualified_columns=fully_qualified_columns,
            num_rows=num_rows,
            use_xmla=use_xmla,
        )
        return self._add_asset(asset)

    @public_api
    def add_powerbi_table_asset(  # noqa: PLR0913 # FIXME CoP
        self,
        name: str,
        table: str,
        batch_metadata: Optional[BatchMetadata] = None,
        fully_qualified_columns: bool = False,
        num_rows: Optional[int] = None,
        multiindex_hierarchies: bool = False,
        mode: Mode = "xmla",
    ) -> PowerBITable:
        """Adds a PowerBITable asset to this datasource.

        Args:
            name: The name of this table asset.
            table_name: The table where the data resides.
            schema: The schema that holds the table.
            batch_metadata: BatchMetadata we want to associate with this DataAsset and all batches derived from it.

        Returns:
            The asset that is added to the datasource.
        """  # noqa: E501 # FIXME CoP
        asset = PowerBITable(
            name=name,
            batch_metadata=batch_metadata or {},
            table=table,
            fully_qualified_columns=fully_qualified_columns,
            num_rows=num_rows,
            multiindex_hierarchies=multiindex_hierarchies,
            mode=mode,
        )
        return self._add_asset(asset)

    @staticmethod
    def _running_on_fabric() -> bool:
        if (
            os.environ.get("AZURE_SERVICE")  # noqa: TID251 # needed for fabric
            != _REQUIRED_FABRIC_SERVICE
        ):
            return False
        from pyspark.sql import SparkSession  # noqa: TID251 # needed for fabric

        sc = SparkSession.builder.getOrCreate().sparkContext
        return sc.getConf().get("spark.cluster.type") != "synapse"



import timeit
import unittest
from unittest.mock import PropertyMock

import pandas as pd
import pytest

from great_expectations.compatibility import aws
from great_expectations.core.expectation_validation_result import (
    ExpectationValidationResult,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.expectations import (
    ExpectColumnValuesToBeInSet,
)
from great_expectations.self_check.util import build_sa_validator_with_data
from great_expectations.util import is_library_loadable


@pytest.mark.skipif(
    not (aws.sqlalchemy_athena and is_library_loadable(library_name="pyathena")),
    reason="pyathena is not installed",
)
@pytest.mark.athena
@pytest.mark.external_sqldialect
def test_expect_column_values_to_be_of_type_string_dialect_pyathena(sa):
    df = pd.DataFrame({"col": ["test_val1", "test_val2"]})
    validator = build_sa_validator_with_data(
        df=df,
        sa_engine_name="sqlite",
        table_name="expect_column_values_to_be_of_type_string_dialect_pyathena_1",
    )

    # Monkey-patch dialect for testing purposes.
    validator.execution_engine.dialect_module = aws.sqlalchemy_athena

    result = validator.expect_column_values_to_be_of_type("col", type_="string")

    assert result == ExpectationValidationResult(
        success=True,
        expectation_config={
            "expectation_type": "expect_column_values_to_be_of_type",
            "kwargs": {
                "column": "col",
                "type_": "string",
            },
            "meta": {},
        },
        result={
            "element_count": 2,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "partial_unexpected_list": [],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.0,
            "unexpected_percent_nonmissing": 0.0,
        },
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        meta={},
    )


@pytest.mark.sqlite
@pytest.mark.external_sqldialect
def test_expect_column_values_to_be_of_type_string_dialect_sqlite(sa):
    df = pd.DataFrame({"col": ["test_val1", "test_val2"]})
    validator = build_sa_validator_with_data(
        df=df,
        sa_engine_name="sqlite",
        table_name="expect_column_values_to_be_of_type_string_dialect_sqlite_1",
    )

    result = validator.expect_column_values_to_be_of_type("col", type_="TEXT")

    assert result == ExpectationValidationResult(
        success=True,
        expectation_config={
            "expectation_type": "expect_column_values_to_be_of_type",
            "kwargs": {
                "column": "col",
                "type_": "TEXT",
            },
            "meta": {},
        },
        result={
            "element_count": 2,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "partial_unexpected_list": [],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.0,
            "unexpected_percent_nonmissing": 0.0,
        },
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        meta={},
    )


@pytest.mark.unit
def test_expect_column_values_to_be_in_set_render_performance():
    """
    This test prevents a historical bug in which rendering took ~10 seconds to render 400 items.
    """

    large_number = 150

    x = ExpectColumnValuesToBeInSet(
        column="foo_column_name", value_set=["foo" for _ in range(large_number)]
    )

    duration_s = timeit.timeit(x.render, number=1)
    assert duration_s < 1, f"Rendering took {duration_s} seconds"


@pytest.mark.unit
@pytest.mark.parametrize(
    "dialect_name",
    [
        GXSqlDialect.DATABRICKS,
        GXSqlDialect.POSTGRESQL,
        GXSqlDialect.SNOWFLAKE,
    ],
)
def test_expect_column_values_to_be_of_type_case_insensitivity(sa, dialect_name):
    df = pd.DataFrame(
        {
            "datetime_col": pd.Series(
                [pd.Timestamp("2023-01-01"), pd.Timestamp("2023-01-02")], dtype="datetime64[ns]"
            ),
            "int_col": pd.Series([1, 2, None], dtype=pd.Int64Dtype()),
            "string_col": pd.Series(["apple", "banana", "cherry"], dtype=pd.StringDtype()),
            "bool_col": pd.Series([True, False, None], dtype=pd.BooleanDtype()),
        }
    )
    validator = build_sa_validator_with_data(
        df=df,
        sa_engine_name="sqlite",  # Using sqlite as base, will mock dialect
        table_name="type_test_table_parametrized",
    )

    # Mock the dialect_name property to test different SQL dialects
    with unittest.mock.patch.object(
        type(validator.execution_engine),
        "dialect_name",
        new_callable=PropertyMock,
        return_value=dialect_name,
    ):
        column_test_configs = [
            {"col_name": "datetime_col", "canonical_type": "DATETIME", "wrong_type": "INTEGER"},
            {"col_name": "int_col", "canonical_type": "BIGINT", "wrong_type": "STRING"},
            # For SQLite, pandas StringDtype typically maps to TEXT
            {"col_name": "string_col", "canonical_type": "TEXT", "wrong_type": "BOOLEAN"},
            {"col_name": "bool_col", "canonical_type": "BOOLEAN", "wrong_type": "DATE"},
        ]

        capitalization_variations = [
            lambda t: t.lower(),
            lambda t: t.upper(),
            lambda t: t.capitalize(),
        ]

        for config in column_test_configs:
            col_name = config["col_name"]
            canonical_type = config["canonical_type"]
            wrong_type = config["wrong_type"]

            # Test correct types with different capitalizations
            for cap_func in capitalization_variations:
                type_str = cap_func(canonical_type)
                result = validator.expect_column_values_to_be_of_type(col_name, type_=type_str)
                assert result.success is True, (
                    f"Expected success=True for type='{type_str}' on column '{col_name}' "
                    f"with dialect '{dialect_name}', "
                    f"but got success={result.success}. "
                    f"Observed value: {result.result.get('observed_value')}"
                )

            # Test incorrect type
            result = validator.expect_column_values_to_be_of_type(col_name, type_=wrong_type)
            assert result.success is False, (
                f"Expected success=False for wrong_type='{wrong_type}' on column '{col_name}' "
                f"with dialect '{dialect_name}', "
                f"but got success={result.success}. "
                f"Observed value: {result.result.get('observed_value')}"
            )



from __future__ import annotations

import enum
from typing import Final, Literal, Union


class ResultFormat(str, enum.Enum):
    BOOLEAN_ONLY = "BOOLEAN_ONLY"
    BASIC = "BASIC"
    COMPLETE = "COMPLETE"
    SUMMARY = "SUMMARY"


ResultFormatUnion = Union[
    ResultFormat, dict, Literal["BOOLEAN_ONLY", "BASIC", "SUMMARY", "COMPLETE"]
]

DEFAULT_RESULT_FORMAT: Final = ResultFormat.SUMMARY



import pandas as pd

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.sample_values import (
    ColumnSampleValues,
    ColumnSampleValuesResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import SQL_DATA_SOURCES

COLUMN_NAME = "stuff"
DATA_FRAME = pd.DataFrame(
    {
        COLUMN_NAME: [
            "a",
            "b",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "a",
            "b",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
            "c",
        ],
    },
)


class TestColumnSampleValues:
    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_column_sample_values_default(self, batch_for_datasource: Batch) -> None:
        metric = ColumnSampleValues(column=COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnSampleValuesResult)
        assert len(metric_result.value) == 20
        assert isinstance(metric_result.value[0], str)

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_column_sample_values_count(self, batch_for_datasource: Batch) -> None:
        metric = ColumnSampleValues(column=COLUMN_NAME, count=10)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnSampleValuesResult)
        assert len(metric_result.value) == 10
        assert isinstance(metric_result.value[0], str)



from __future__ import annotations

import datetime
import os
from typing import List
from unittest import mock

import pandas as pd
import pytest
from dateutil.parser import parse

from great_expectations.core.batch_spec import SqlAlchemyDatasourceBatchSpec
from great_expectations.data_context.util import file_relative_path
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_partitioner import (
    DatePart,
    SqlAlchemyDataPartitioner,
)
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from great_expectations.self_check.util import build_sa_execution_engine
from tests.execution_engine.partition_and_sample.partition_and_sample_test_cases import (
    MULTIPLE_DATE_PART_BATCH_IDENTIFIERS,
    MULTIPLE_DATE_PART_DATE_PARTS,
    SINGLE_DATE_PART_BATCH_IDENTIFIERS,
    SINGLE_DATE_PART_DATE_PARTS,
)

# Here we add SqlAlchemyDataPartitioner specific test cases to the generic test cases:
from tests.integration.fixtures.partition_and_sample_data.partitioner_test_cases_and_fixtures import (  # noqa: E501 # FIXME CoP
    TaxiPartitioningTestCase,
    TaxiPartitioningTestCasesBase,
    TaxiPartitioningTestCasesDateTime,
    TaxiPartitioningTestCasesWholeTable,
    TaxiTestData,
)
from tests.test_utils import convert_string_columns_to_datetime

SINGLE_DATE_PART_DATE_PARTS += [
    pytest.param(
        [SqlAlchemyDataPartitioner.date_part.MONTH],
        id="month getting date parts from SqlAlchemyDataPartitioner.date_part",
    )
]
MULTIPLE_DATE_PART_DATE_PARTS += [
    pytest.param(
        [
            SqlAlchemyDataPartitioner.date_part.YEAR,
            SqlAlchemyDataPartitioner.date_part.MONTH,
        ],
        id="year_month getting date parts from SqlAlchemyDataPartitioner.date_part",
    )
]

pytestmark = [
    pytest.mark.sqlalchemy_version_compatibility,
    pytest.mark.external_sqldialect,
]


@mock.patch(
    "great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_partitioner.SqlAlchemyDataPartitioner.partition_on_date_parts"
)
@pytest.mark.parametrize(
    "partitioner_method_name,called_with_date_parts",
    [
        ("partition_on_year", [DatePart.YEAR]),
        ("partition_on_year_and_month", [DatePart.YEAR, DatePart.MONTH]),
        (
            "partition_on_year_and_month_and_day",
            [DatePart.YEAR, DatePart.MONTH, DatePart.DAY],
        ),
    ],
)
@pytest.mark.sqlite
def test_named_date_part_methods(
    mock_partition_on_date_parts: mock.MagicMock,  # noqa: TID251 # FIXME CoP
    partitioner_method_name: str,
    called_with_date_parts: List[DatePart],
):
    """Test that a partially pre-filled version of partition_on_date_parts() was called with the appropriate params.
    For example, partition_on_year.
    """  # noqa: E501 # FIXME CoP
    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    column_name: str = "column_name"
    batch_identifiers: dict = {column_name: {"year": 2018, "month": 10, "day": 31}}

    getattr(data_partitioner, partitioner_method_name)(
        column_name=column_name,
        batch_identifiers=batch_identifiers,
    )

    mock_partition_on_date_parts.assert_called_with(
        column_name=column_name,
        batch_identifiers=batch_identifiers,
        date_parts=called_with_date_parts,
    )


@pytest.mark.parametrize(
    "batch_identifiers_for_column",
    SINGLE_DATE_PART_BATCH_IDENTIFIERS,
)
@pytest.mark.parametrize(
    "date_parts",
    SINGLE_DATE_PART_DATE_PARTS,
)
@pytest.mark.sqlite
def test_partition_on_date_parts_single_date_parts(batch_identifiers_for_column, date_parts, sa):
    """What does this test and why?

    partition_on_date_parts should still build the correct query when passed a single element list
     date_parts that is a string, DatePart enum objects, mixed case string.
     To match our interface it should accept a dateutil parseable string as the batch identifier
     or a datetime and also fail when parameters are invalid.
    """

    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    column_name: str = "column_name"

    result: sa.sql.elements.BooleanClauseList = data_partitioner.partition_on_date_parts(
        column_name=column_name,
        batch_identifiers={column_name: batch_identifiers_for_column},
        date_parts=date_parts,
    )

    # using values
    assert isinstance(result, sa.sql.elements.BinaryExpression)
    assert isinstance(result.comparator.type, sa.Boolean)
    assert isinstance(result.left, sa.sql.elements.Extract)
    assert result.left.field == "month"
    assert result.left.expr.name == column_name
    assert result.right.effective_value == 10


@pytest.mark.parametrize(
    "batch_identifiers_for_column",
    MULTIPLE_DATE_PART_BATCH_IDENTIFIERS,
)
@pytest.mark.parametrize(
    "date_parts",
    MULTIPLE_DATE_PART_DATE_PARTS,
)
@pytest.mark.sqlite
def test_partition_on_date_parts_multiple_date_parts(batch_identifiers_for_column, date_parts, sa):
    """What does this test and why?

    partition_on_date_parts should still build the correct query when passed
     date parts that are strings, DatePart enum objects, a mixture and mixed case.
     To match our interface it should accept a dateutil parseable string as the batch identifier
     or a datetime and also fail when parameters are invalid.
    """

    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    column_name: str = "column_name"

    result: sa.sql.elements.BooleanClauseList = data_partitioner.partition_on_date_parts(
        column_name=column_name,
        batch_identifiers={column_name: batch_identifiers_for_column},
        date_parts=date_parts,
    )

    # using values
    assert isinstance(result, sa.sql.elements.BooleanClauseList)

    assert isinstance(result.clauses[0].comparator.type, sa.Boolean)
    assert isinstance(result.clauses[0].left, sa.sql.elements.Extract)
    assert result.clauses[0].left.field == "year"
    assert result.clauses[0].left.expr.name == column_name
    assert result.clauses[0].right.effective_value == 2018

    assert isinstance(result.clauses[1].comparator.type, sa.Boolean)
    assert isinstance(result.clauses[1].left, sa.sql.elements.Extract)
    assert result.clauses[1].left.field == "month"
    assert result.clauses[1].left.expr.name == column_name
    assert result.clauses[1].right.effective_value == 10


@mock.patch(
    "great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_partitioner.SqlAlchemyDataPartitioner.get_data_for_batch_identifiers_for_partition_on_date_parts"
)
@mock.patch("great_expectations.execution_engine.execution_engine.ExecutionEngine")
@pytest.mark.sqlite
def test_get_data_for_batch_identifiers_year(
    mock_execution_engine: mock.MagicMock,  # noqa: TID251 # FIXME CoP
    mock_get_data_for_batch_identifiers_for_partition_on_date_parts: mock.MagicMock,  # noqa: TID251 # FIXME CoP
):
    """test that get_data_for_batch_identifiers_for_partition_on_date_parts() was called with the appropriate params."""  # noqa: E501 # FIXME CoP
    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    # selectable should be a sa.Selectable object but since we are mocking out
    # get_data_for_batch_identifiers_for_partition_on_date_parts
    # and just verifying its getting passed through, we ignore the type here.
    selectable: str = "mock_selectable"
    column_name: str = "column_name"

    data_partitioner.get_data_for_batch_identifiers_year(
        execution_engine=mock_execution_engine,
        selectable=selectable,
        column_name=column_name,
    )

    mock_get_data_for_batch_identifiers_for_partition_on_date_parts.assert_called_with(
        execution_engine=mock_execution_engine,
        selectable=selectable,
        column_name=column_name,
        date_parts=[DatePart.YEAR],
    )


@mock.patch(
    "great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_partitioner.SqlAlchemyDataPartitioner.get_data_for_batch_identifiers_for_partition_on_date_parts"
)
@mock.patch("great_expectations.execution_engine.execution_engine.ExecutionEngine")
@pytest.mark.sqlite
def test_get_data_for_batch_identifiers_year_and_month(
    mock_execution_engine: mock.MagicMock,  # noqa: TID251 # FIXME CoP
    mock_get_data_for_batch_identifiers_for_partition_on_date_parts: mock.MagicMock,  # noqa: TID251 # FIXME CoP
):
    """test that get_data_for_batch_identifiers_for_partition_on_date_parts() was called with the appropriate params."""  # noqa: E501 # FIXME CoP
    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    selectable: str = "mock_selectable"
    column_name: str = "column_name"

    data_partitioner.get_data_for_batch_identifiers_year_and_month(
        execution_engine=mock_execution_engine,
        selectable=selectable,
        column_name=column_name,
    )

    mock_get_data_for_batch_identifiers_for_partition_on_date_parts.assert_called_with(
        execution_engine=mock_execution_engine,
        selectable=selectable,
        column_name=column_name,
        date_parts=[DatePart.YEAR, DatePart.MONTH],
    )


@mock.patch(
    "great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_partitioner.SqlAlchemyDataPartitioner.get_data_for_batch_identifiers_for_partition_on_date_parts"
)
@mock.patch("great_expectations.execution_engine.execution_engine.ExecutionEngine")
@pytest.mark.sqlite
def test_get_data_for_batch_identifiers_year_and_month_and_day(
    mock_execution_engine: mock.MagicMock,  # noqa: TID251 # FIXME CoP
    mock_get_data_for_batch_identifiers_for_partition_on_date_parts: mock.MagicMock,  # noqa: TID251 # FIXME CoP
):
    """test that get_data_for_batch_identifiers_for_partition_on_date_parts() was called with the appropriate params."""  # noqa: E501 # FIXME CoP
    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    selectable: str = "mock_selectable"
    column_name: str = "column_name"

    data_partitioner.get_data_for_batch_identifiers_year_and_month_and_day(
        execution_engine=mock_execution_engine,
        selectable=selectable,
        column_name=column_name,
    )

    mock_get_data_for_batch_identifiers_for_partition_on_date_parts.assert_called_with(
        execution_engine=mock_execution_engine,
        selectable=selectable,
        column_name=column_name,
        date_parts=[DatePart.YEAR, DatePart.MONTH, DatePart.DAY],
    )


@pytest.mark.parametrize(
    "date_parts",
    SINGLE_DATE_PART_DATE_PARTS,
)
@pytest.mark.sqlite
def test_get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts_single_date_parts(  # noqa: E501 # FIXME CoP
    date_parts, sa
):
    """What does this test and why?

    get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts should still build the correct
    query when passed a single element list of date_parts that is a string, DatePart enum object, or mixed case string.
    """  # noqa: E501 # FIXME CoP

    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")
    selectable: sa.sql.Selectable = sa.text("table_name")
    column_name: str = "column_name"

    result: sa.sql.elements.BooleanClauseList = data_partitioner.get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts(  # noqa: E501 # FIXME CoP
        selectable=selectable,
        column_name=column_name,
        date_parts=date_parts,
    )

    assert isinstance(result, sa.sql.Select)

    query_str: str = (
        str(result.compile(compile_kwargs={"literal_binds": True}))
        .replace("\n", "")
        .replace(" ", "")
        .lower()
    )
    assert (
        query_str
        == (
            "SELECT distinct(EXTRACT(month FROM column_name)) AS concat_distinct_values, "
            f"CAST(EXTRACT(month FROM column_name) AS INTEGER) AS month FROM {selectable}"
        )
        .replace("\n", "")
        .replace(" ", "")
        .lower()
    )


@pytest.mark.parametrize(
    "date_parts",
    MULTIPLE_DATE_PART_DATE_PARTS,
)
@pytest.mark.parametrize(
    "dialect,expected_query_str",
    [
        pytest.param(
            "sqlite",
            "SELECT DISTINCT(CAST(EXTRACT(year FROM column_name) AS VARCHAR) || CAST (EXTRACT(month FROM column_name) AS VARCHAR)) AS concat_distinct_values, CAST(EXTRACT(year FROM column_name) AS INTEGER) AS year, CAST(EXTRACT(month FROM column_name) AS INTEGER) AS month FROM table_name",  # noqa: E501 # FIXME CoP
            marks=pytest.mark.sqlite,
            id="sqlite",
        ),
        pytest.param(
            "postgres",
            "SELECT DISTINCT(CONCAT(CONCAT('', CAST(EXTRACT(year FROM column_name) AS VARCHAR)), CAST(EXTRACT(month FROM column_name) AS VARCHAR))) AS concat_distinct_values, CAST(EXTRACT(year FROM column_name) AS INTEGER) AS year, CAST(EXTRACT(month FROM column_name) AS INTEGER) AS month FROM table_name",  # noqa: E501 # FIXME CoP
            marks=pytest.mark.postgresql,
            id="postgres",
        ),
    ],
)
def test_get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts_multiple_date_parts(  # noqa: E501 # FIXME CoP
    date_parts, dialect, expected_query_str, sa
):
    """What does this test and why?
    get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts should
    return the correct query when passed any valid set of parameters including multiple date parts.
    """
    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect=dialect)
    selectable: sa.sql.Selectable = sa.text("table_name")
    column_name: str = "column_name"

    result: sa.sql.elements.BooleanClauseList = data_partitioner.get_partition_query_for_data_for_batch_identifiers_for_partition_on_date_parts(  # noqa: E501 # FIXME CoP
        selectable=selectable,
        column_name=column_name,
        date_parts=date_parts,
    )

    assert isinstance(result, sa.sql.Select)

    actual_query_str: str = (
        str(result.compile(compile_kwargs={"literal_binds": True}))
        .replace("\n", "")
        .replace(" ", "")
        .lower()
    )
    assert actual_query_str == expected_query_str.replace("\n", "").replace(" ", "").lower()


@pytest.mark.parametrize(
    "underscore_prefix",
    [
        pytest.param("_", id="underscore prefix"),
        pytest.param("", id="no underscore prefix"),
    ],
)
@pytest.mark.parametrize(
    "partitioner_method_name",
    [
        pytest.param(partitioner_method_name, id=partitioner_method_name)
        for partitioner_method_name in [
            "partition_on_year",
            "partition_on_year_and_month",
            "partition_on_year_and_month_and_day",
            "partition_on_date_parts",
            "partition_on_whole_table",
            "partition_on_column_value",
            "partition_on_converted_datetime",
            "partition_on_divided_integer",
            "partition_on_mod_integer",
            "partition_on_multi_column_values",
            "partition_on_hashed_column",
        ]
    ],
)
@pytest.mark.sqlite
def test_get_partitioner_method(underscore_prefix: str, partitioner_method_name: str):
    data_partitioner: SqlAlchemyDataPartitioner = SqlAlchemyDataPartitioner(dialect="sqlite")

    partitioner_method_name_with_prefix = f"{underscore_prefix}{partitioner_method_name}"

    assert data_partitioner.get_partitioner_method(partitioner_method_name_with_prefix) == getattr(
        data_partitioner, partitioner_method_name
    )


def ten_trips_per_month_df() -> pd.DataFrame:
    csv_path: str = file_relative_path(
        os.path.dirname(os.path.dirname(__file__)),  # noqa: PTH120 # FIXME CoP
        os.path.join(  # noqa: PTH118 # FIXME CoP
            "test_sets",
            "taxi_yellow_tripdata_samples",
            "ten_trips_from_each_month",
            "yellow_tripdata_sample_10_trips_from_each_month.csv",
        ),
    )
    df: pd.DataFrame = pd.read_csv(csv_path)
    return df


@pytest.fixture
def in_memory_sqlite_taxi_ten_trips_per_month_execution_engine(sa):
    df: pd.DataFrame = ten_trips_per_month_df()
    convert_string_columns_to_datetime(
        df=df, column_names_to_convert=["pickup_datetime", "dropoff_datetime"]
    )
    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df, sa)
    return engine


@pytest.mark.parametrize(
    "taxi_test_cases",
    [
        TaxiPartitioningTestCasesWholeTable(
            taxi_test_data=TaxiTestData(
                test_df=ten_trips_per_month_df(),
                test_column_name=None,
                test_column_names=None,
                column_names_to_convert=["pickup_datetime", "dropoff_datetime"],
            )
        ),
        TaxiPartitioningTestCasesDateTime(
            taxi_test_data=TaxiTestData(
                test_df=ten_trips_per_month_df(),
                test_column_name="pickup_datetime",
                test_column_names=None,
                column_names_to_convert=["pickup_datetime", "dropoff_datetime"],
            )
        ),
    ],
)
@pytest.mark.xfail(reason="To be implemented in V1-305", strict=True)
@pytest.mark.sqlite
def test_sqlite_partition(
    taxi_test_cases: TaxiPartitioningTestCasesBase,
    sa,
):
    """What does this test and why?
    partitioners should work with sqlite.
    """
    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(taxi_test_cases.test_df, sa)

    test_cases: List[TaxiPartitioningTestCase] = taxi_test_cases.test_cases()
    test_case: TaxiPartitioningTestCase
    batch_spec: SqlAlchemyDatasourceBatchSpec
    for test_case in test_cases:
        if test_case.table_domain_test_case:
            batch_spec = SqlAlchemyDatasourceBatchSpec(
                table_name="test",
                schema_name="main",
                partitioner_method=test_case.add_batch_definition_method_name,
                partitioner_kwargs=test_case.add_batch_definition_kwargs,
                batch_identifiers={},
            )
        else:  # noqa: PLR5501 # FIXME CoP
            if taxi_test_cases.test_column_name:
                assert test_case.expected_column_values is not None
                batch_spec = SqlAlchemyDatasourceBatchSpec(
                    table_name="test",
                    schema_name="main",
                    partitioner_method=test_case.add_batch_definition_method_name,
                    partitioner_kwargs=test_case.add_batch_definition_kwargs,
                    batch_identifiers={
                        taxi_test_cases.test_column_name: test_case.expected_column_values[0]
                    },
                )
            elif taxi_test_cases.test_column_names:
                assert test_case.expected_column_values is not None
                column_name: str
                batch_spec = SqlAlchemyDatasourceBatchSpec(
                    table_name="test",
                    schema_name="main",
                    partitioner_method=test_case.add_batch_definition_method_name,
                    partitioner_kwargs=test_case.add_batch_definition_kwargs,
                    batch_identifiers={
                        column_name: test_case.expected_column_values[0][column_name]
                        for column_name in taxi_test_cases.test_column_names
                    },
                )
            else:
                raise ValueError("Missing test_column_names or test_column_names attribute.")

        batch_data: SqlAlchemyBatchData = engine.get_batch_data(batch_spec=batch_spec)

        # Right number of rows?
        num_rows: int = batch_data.execution_engine.execute_query(
            sa.select(sa.func.count()).select_from(batch_data.selectable)
        ).scalar()
        # noinspection PyUnresolvedReferences
        assert num_rows == test_case.num_expected_rows_in_first_batch_definition


@pytest.mark.sqlite
def test_sqlite_partition_on_year(sa, in_memory_sqlite_taxi_ten_trips_per_month_execution_engine):
    """What does this test and why?
    partitioners should work with sqlite and return the correct rows.
    """

    engine: SqlAlchemyExecutionEngine = in_memory_sqlite_taxi_ten_trips_per_month_execution_engine

    n: int = 120
    batch_spec: SqlAlchemyDatasourceBatchSpec = SqlAlchemyDatasourceBatchSpec(
        table_name="test",
        schema_name="main",
        partitioner_method="partition_on_year",
        partitioner_kwargs={"column_name": "pickup_datetime"},
        batch_identifiers={"pickup_datetime": "2018"},
    )
    batch_data: SqlAlchemyBatchData = engine.get_batch_data(batch_spec=batch_spec)

    # Right number of rows?
    num_rows: int = batch_data.execution_engine.execute_query(
        sa.select(sa.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == n

    # Right rows?
    rows: list[sa.RowMapping] = (
        batch_data.execution_engine.execute_query(
            sa.select(sa.text("*")).select_from(batch_data.selectable)
        )
        .mappings()
        .fetchall()
    )

    row_dates: List[datetime.datetime] = [parse(row["pickup_datetime"]) for row in rows]
    for row_date in row_dates:
        assert row_date.month >= 1
        assert row_date.month <= 12
        assert row_date.year == 2018


@pytest.mark.sqlite
def test_sqlite_partition_and_sample_using_limit(
    sa, in_memory_sqlite_taxi_ten_trips_per_month_execution_engine
):
    """What does this test and why?
    partitioners and samplers should work together in sqlite.
    """

    engine: SqlAlchemyExecutionEngine = in_memory_sqlite_taxi_ten_trips_per_month_execution_engine

    n: int = 3
    batch_spec: SqlAlchemyDatasourceBatchSpec = SqlAlchemyDatasourceBatchSpec(
        table_name="test",
        schema_name="main",
        sampling_method="sample_using_limit",
        sampling_kwargs={"n": n},
        partitioner_method="partition_on_year",
        partitioner_kwargs={"column_name": "pickup_datetime"},
        batch_identifiers={"pickup_datetime": "2018"},
    )
    batch_data: SqlAlchemyBatchData = engine.get_batch_data(batch_spec=batch_spec)

    # Right number of rows?
    num_rows: int = batch_data.execution_engine.execute_query(
        sa.select(sa.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == n

    # Right rows?
    rows: list[sa.RowMapping] = (
        batch_data.execution_engine.execute_query(
            sa.select(sa.text("*")).select_from(batch_data.selectable)
        )
        .mappings()
        .fetchall()
    )

    row_dates: List[datetime.datetime] = [parse(row["pickup_datetime"]) for row in rows]
    for row_date in row_dates:
        assert row_date.month == 1
        assert row_date.year == 2018



from __future__ import annotations

import copy
import datetime
import logging
import math
import operator
import traceback
from collections import namedtuple
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

import dateutil
from pyparsing import (
    CaselessKeyword,
    Forward,
    Group,
    Literal,
    ParseException,
    ParseResults,
    Regex,
    Suppress,
    Word,
    alphanums,
    alphas,
    delimitedList,
    dictOf,
)

from great_expectations.exceptions import SuiteParameterError
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP

if TYPE_CHECKING:
    from typing_extensions import TypeAlias, TypeGuard

    from great_expectations.data_context import AbstractDataContext

logger = logging.getLogger(__name__)
_epsilon = 1e-12

# NOTE: Temporary alias - to be converted to a rich type
SuiteParameterDict: TypeAlias = dict


def is_suite_parameter(value: Any) -> TypeGuard[SuiteParameterDict]:
    """Typeguard to check if a value is an suite parameter."""
    return isinstance(value, dict) and "$PARAMETER" in value


def get_suite_parameter_key(suite_parameter: SuiteParameterDict) -> str:
    """Get the key of a suite parameter.

    e.g. if the suite parameter is {"$PARAMETER": "foo"}, this function will return "foo".
    When evaluating the runtime configuration of an expectation, we will look for
    a runtime value for "foo".

    Args:
        suite_parameter: The suite parameter to get the key of

    Returns:
        The key of the suite parameter
    """
    return suite_parameter["$PARAMETER"]


class SuiteParameterParser:
    """
    This Suite Parameter Parser uses pyparsing to provide a basic expression language capable of evaluating
    parameters using values available only at run time.

    expop   :: '^'
    multop  :: '*' | '/'
    addop   :: '+' | '-'
    integer :: ['+' | '-'] '0'..'9'+
    atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
    factor  :: atom [ expop factor ]*
    term    :: factor [ multop factor ]*
    expr    :: term [ addop term ]*

    The parser is modified from: https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py
    """  # noqa: E501 # FIXME CoP

    # map operator symbols to corresponding arithmetic operations
    opn = {
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,
        "/": operator.truediv,
        "^": operator.pow,
    }

    fn = {
        "sin": math.sin,
        "cos": math.cos,
        "tan": math.tan,
        "exp": math.exp,
        "abs": abs,
        "trunc": lambda a: int(a),
        "round": round,
        "sgn": lambda a: -1 if a < -_epsilon else 1 if a > _epsilon else 0,
        "now": datetime.datetime.now,
        "datetime": datetime.datetime,
        "timedelta": datetime.timedelta,
    }

    def __init__(self) -> None:
        self.exprStack: list = []
        self._parser = None

    def push_first(self, toks) -> None:
        self.exprStack.append(toks[0])

    def push_unary_minus(self, toks) -> None:
        for t in toks:
            if t == "-":
                self.exprStack.append("unary -")
            else:
                break

    def clear_stack(self) -> None:
        del self.exprStack[:]

    def get_parser(self):
        self.clear_stack()
        if not self._parser:
            # use CaselessKeyword for e and pi, to avoid accidentally matching
            # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
            # and CaselessKeyword only match whole words
            e = CaselessKeyword("E")
            pi = CaselessKeyword("PI")
            # fnumber = Combine(Word("+-"+nums, nums) +
            #                    Optional("." + Optional(Word(nums))) +
            #                    Optional(e + Word("+-"+nums, nums)))
            # or use provided pyparsing_common.number, but convert back to str:
            # fnumber = ppc.number().addParseAction(lambda t: str(t[0]))
            fnumber = Regex(r"[+-]?(?:\d+|\.\d+)(?:\.\d+)?(?:[eE][+-]?\d+)?")
            variable = Word(alphas, f"{alphanums}_$")

            plus, minus, mult, div = map(Literal, "+-*/")
            lpar, rpar = map(Suppress, "()")
            addop = plus | minus
            multop = mult | div
            expop = Literal("^")

            expr = Forward()
            expr_list = delimitedList(Group(expr))

            # We will allow functions either to accept *only* keyword
            # expressions or *only* non-keyword expressions
            # define function keyword arguments
            key = Word(f"{alphas}_") + Suppress("=")
            # value = (fnumber | Word(alphanums))
            value = expr
            keyval = dictOf(key.setParseAction(self.push_first), value)
            kwarglist = delimitedList(keyval)

            # add parse action that replaces the function identifier with a (name, number of args, has_fn_kwargs) tuple  # noqa: E501 # FIXME CoP
            # 20211009 - JPC - Note that it's important that we consider kwarglist
            # first as part of disabling backtracking for the function's arguments
            fn_call = (variable + lpar + rpar).setParseAction(
                lambda t: t.insert(0, (t.pop(0), 0, False))
            ) | (
                (variable + lpar - Group(expr_list) + rpar).setParseAction(
                    lambda t: t.insert(0, (t.pop(0), len(t[0]), False))
                )
                ^ (variable + lpar - Group(kwarglist) + rpar).setParseAction(
                    lambda t: t.insert(0, (t.pop(0), len(t[0]), True))
                )
            )
            atom = (
                addop[...]
                + (
                    (fn_call | pi | e | fnumber | variable).setParseAction(self.push_first)
                    | Group(lpar + expr + rpar)
                )
            ).setParseAction(self.push_unary_minus)

            # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left  # noqa: E501 # FIXME CoP
            # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
            factor = Forward()
            factor <<= atom + (expop + factor).setParseAction(self.push_first)[...]
            term = factor + (multop + factor).setParseAction(self.push_first)[...]
            expr <<= term + (addop + term).setParseAction(self.push_first)[...]
            self._parser = expr
        return self._parser

    def evaluate_stack(self, s):  # noqa: C901, PLR0911, PLR0912 # FIXME CoP
        op, num_args, has_fn_kwargs = s.pop(), 0, False
        if isinstance(op, tuple):
            op, num_args, has_fn_kwargs = op
        if op == "unary -":
            return -self.evaluate_stack(s)
        if op in "+-*/^":
            # note: operands are pushed onto the stack in reverse order
            op2 = self.evaluate_stack(s)
            op1 = self.evaluate_stack(s)
            return self.opn[op](op1, op2)
        elif op == "PI":
            return math.pi  # 3.1415926535
        elif op == "E":
            return math.e  # 2.718281828
        elif op in self.fn:
            # note: args are pushed onto the stack in reverse order
            if has_fn_kwargs:
                kwargs = dict()
                for _ in range(num_args):
                    v = self.evaluate_stack(s)
                    k = s.pop()
                    kwargs.update({k: v})
                return self.fn[op](**kwargs)
            else:
                args = reversed([self.evaluate_stack(s) for _ in range(num_args)])
                return self.fn[op](*args)
        else:
            # Require that the *entire* expression evaluates to number or datetime UNLESS there is *exactly one*  # noqa: E501 # FIXME CoP
            # expression to substitute (see cases where len(parse_results) == 1 in the parse_suite_parameter  # noqa: E501 # FIXME CoP
            # method).
            evaluated: Union[int, float, datetime.datetime]
            try:
                evaluated = int(op)
                logger.info("Suite parameter operand successfully parsed as integer.")
            except ValueError:
                logger.info("Parsing suite parameter operand as integer failed.")
                try:
                    evaluated = float(op)
                    logger.info("Suite parameter operand successfully parsed as float.")
                except ValueError:
                    logger.info("Parsing suite parameter operand as float failed.")
                    try:
                        evaluated = dateutil.parser.parse(op)
                        logger.info("Suite parameter operand successfully parsed as datetime.")
                    except ValueError as e:
                        logger.info("Parsing suite parameter operand as datetime failed.")
                        raise e  # noqa: TRY201 # FIXME CoP
            return evaluated


def build_suite_parameters(
    expectation_args: dict,
    suite_parameters: Optional[dict] = None,
    interactive_evaluation: bool = True,
    data_context=None,
) -> Tuple[dict, dict]:
    """Build a dictionary of parameters to evaluate, using the provided suite_parameters,
    AND mutate expectation_args by removing any parameter values passed in as temporary values during
    exploratory work.
    """  # noqa: E501 # FIXME CoP
    suite_args = copy.deepcopy(expectation_args)
    substituted_parameters = {}

    # Iterate over arguments, and replace $PARAMETER-defined args with their
    # specified parameters.
    for key, value in suite_args.items():
        if isinstance(value, dict) and "$PARAMETER" in value:
            # We do not even need to search for a value if we are not going to do interactive evaluation  # noqa: E501 # FIXME CoP
            if not interactive_evaluation:
                continue

            # First, check to see whether an argument was supplied at runtime
            # If it was, use that one, but remove it from the stored config
            param_key = f"$PARAMETER.{value['$PARAMETER']}"
            if param_key in value:
                suite_args[key] = suite_args[key][param_key]
                del expectation_args[key][param_key]

            # If not, try to parse the suite parameter and substitute, which will raise
            # an exception if we do not have a value
            else:
                raw_value = value["$PARAMETER"]
                parameter_value = parse_suite_parameter(
                    raw_value,
                    suite_parameters=suite_parameters,
                    data_context=data_context,
                )
                suite_args[key] = parameter_value
                # Once we've substituted, we also track that we did so
                substituted_parameters[key] = parameter_value

    return suite_args, substituted_parameters


EXPR = SuiteParameterParser()


def parse_suite_parameter(  # noqa: C901 # FIXME CoP
    parameter_expression: str,
    suite_parameters: Optional[Dict[str, Any]] = None,
    data_context: Optional[AbstractDataContext] = None,
) -> Any:
    """Use the provided suite_parameters dict to parse a given parameter expression.

    Args:
        parameter_expression (str): A string, potentially containing basic arithmetic operations and functions,
            and variables to be substituted
        suite_parameters (dict): A dictionary of name-value pairs consisting of values to substitute
        data_context (DataContext): A data context to use to obtain metrics, if necessary

    The parser will allow arithmetic operations +, -, /, *, as well as basic functions, including trunc() and round() to
    obtain integer values when needed for certain expectations (e.g. expect_column_value_length_to_be_between).

    Valid variables must begin with an alphabetic character and may contain alphanumeric characters plus '_' and '$'.
    """  # noqa: E501 # FIXME CoP
    if suite_parameters is None:
        suite_parameters = {}

    parse_results: Union[ParseResults, list] = _get_parse_results(parameter_expression)

    if _is_single_function_no_args(parse_results):
        # Necessary to catch `now()` (which only needs to be evaluated with `expr.exprStack`)
        # NOTE: 20211122 - Chetan - Any future built-ins that are zero arity functions will match this behavior  # noqa: E501 # FIXME CoP
        pass

    elif len(parse_results) == 1 and parse_results[0] not in suite_parameters:
        # In this special case there were no operations to find, so only one value, but we don't have something to  # noqa: E501 # FIXME CoP
        # substitute for that value
        raise SuiteParameterError(  # noqa: TRY003 # FIXME CoP
            f"No value found for $PARAMETER {parse_results[0]!s}"
        )

    elif len(parse_results) == 1:
        # In this case, we *do* have a substitution for a single type. We treat this specially because in this  # noqa: E501 # FIXME CoP
        # case, we allow complex type substitutions (i.e. do not coerce to string as part of parsing)  # noqa: E501 # FIXME CoP
        # NOTE: 20201023 - JPC - to support MetricDefinition as an suite parameter type, we need to handle that  # noqa: E501 # FIXME CoP
        # case here; is the suite parameter provided here in fact a metric definition?
        return suite_parameters[parse_results[0]]

    elif len(parse_results) == 0 or parse_results[0] != "Parse Failure":
        # we have a stack to evaluate and there was no parse failure.
        for i, ob in enumerate(EXPR.exprStack):
            if isinstance(ob, str) and ob in suite_parameters:
                EXPR.exprStack[i] = str(suite_parameters[ob])

    else:
        err_str, err_line, err_col = parse_results[-1]
        raise SuiteParameterError(  # noqa: TRY003 # FIXME CoP
            f"Parse Failure: {err_str}\nStatement: {err_line}\nColumn: {err_col}"
        )

    try:
        result = EXPR.evaluate_stack(EXPR.exprStack)
        result = convert_to_json_serializable(result)
    except Exception as e:
        exception_traceback = traceback.format_exc()
        exception_message = f'{type(e).__name__}: "{e!s}".  Traceback: "{exception_traceback}".'
        logger.debug(exception_message, e, exc_info=True)
        raise SuiteParameterError(  # noqa: TRY003 # FIXME CoP
            f"Error while evaluating suite parameter expression: {e!s}"
        ) from e

    return result


def _get_parse_results(
    parameter_expression: str,
) -> Union[ParseResults, list]:
    # Calling get_parser clears the stack
    parser = EXPR.get_parser()
    try:
        parse_results = parser.parseString(parameter_expression, parseAll=True)
    except ParseException as err:
        parse_results = [
            "Parse Failure",
            parameter_expression,
            (str(err), err.line, err.column),
        ]
    return parse_results


def _is_single_function_no_args(parse_results: Union[ParseResults, list]) -> bool:
    # Represents a valid parser result of a single function that has no arguments
    return (
        len(parse_results) == 1
        and isinstance(parse_results[0], tuple)
        and parse_results[0][2] is False
    )


def _deduplicate_suite_parameter_dependencies(dependencies: dict) -> dict:  # noqa: C901 #  too complex
    deduplicated: dict = {}
    for suite_name, required_metrics in dependencies.items():
        deduplicated[suite_name] = []
        metrics = set()
        metric_kwargs: dict = {}
        for metric in required_metrics:
            if isinstance(metric, str):
                metrics.add(metric)
            elif isinstance(metric, dict):
                # There is a single metric_kwargs_id object in this construction
                for kwargs_id, metric_list in metric["metric_kwargs_id"].items():
                    if kwargs_id not in metric_kwargs:
                        metric_kwargs[kwargs_id] = set()
                    for metric_name in metric_list:
                        metric_kwargs[kwargs_id].add(metric_name)
        deduplicated[suite_name] = list(metrics)
        if len(metric_kwargs) > 0:
            deduplicated[suite_name] = deduplicated[suite_name] + [
                {
                    "metric_kwargs_id": {
                        metric_kwargs: list(metrics_set)
                        for (metric_kwargs, metrics_set) in metric_kwargs.items()
                    }
                }
            ]

    return deduplicated


SuiteParameterIdentifier = namedtuple(  # noqa: PYI024 # this class is not used
    "SuiteParameterIdentifier",
    ["expectation_suite_name", "metric_name", "metric_kwargs_id"],
)



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

INTEGER_COLUMN = "integers"
INTEGER_AND_NULL_COLUMN = "integers_and_nulls"
STRING_COLUMN = "strings"
NULL_COLUMN = "nulls"


DATA = pd.DataFrame(
    {
        INTEGER_COLUMN: [1, 2, 3, 4, 5],
        INTEGER_AND_NULL_COLUMN: [1, 2, 3, 4, None],
        STRING_COLUMN: ["a", "b", "c", "d", "e"],
        NULL_COLUMN: pd.Series([None, None, None, None, None]),
    },
    dtype="object",
)


try:
    from great_expectations.compatibility.pyspark import types as PYSPARK_TYPES

    SPARK_COLUMN_TYPES = {
        INTEGER_COLUMN: PYSPARK_TYPES.IntegerType,
        INTEGER_AND_NULL_COLUMN: PYSPARK_TYPES.IntegerType,
        STRING_COLUMN: PYSPARK_TYPES.StringType,
        NULL_COLUMN: PYSPARK_TYPES.IntegerType,
    }
except ModuleNotFoundError:
    SPARK_COLUMN_TYPES = {}


@parameterize_batch_for_data_sources(
    data_source_configs=[
        PandasDataFrameDatasourceTestConfig(),
        PandasFilesystemCsvDatasourceTestConfig(),
    ],
    data=DATA,
)
def test_success_for_type__int(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="int")
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[
        BigQueryDatasourceTestConfig(),
        MSSQLDatasourceTestConfig(),
        MySQLDatasourceTestConfig(),
        PostgreSQLDatasourceTestConfig(),
        RedshiftDatasourceTestConfig(),
        SqliteDatasourceTestConfig(),
    ],
    data=DATA,
)
def test_success_for_type__INTEGER(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="INTEGER")
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[DatabricksDatasourceTestConfig()],
    data=DATA,
)
def test_success_for_type__Integer(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="INT")
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[
        SparkFilesystemCsvDatasourceTestConfig(
            column_types=SPARK_COLUMN_TYPES,
        )
    ],
    data=DATA,
)
def test_success_for_type__IntegerType(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="IntegerType")
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[SnowflakeDatasourceTestConfig()],
    data=DATA,
)
def test_success_for_type__Number(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="DECIMAL(38, 0)")
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="int"),
        gxe.ExpectColumnValuesToBeOfType(column=INTEGER_AND_NULL_COLUMN, type_="int"),
        gxe.ExpectColumnValuesToBeOfType(column=STRING_COLUMN, type_="str"),
        gxe.ExpectColumnValuesToBeOfType(column=NULL_COLUMN, type_="float64"),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES,
    data=DATA,
)
def test_success_types(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValuesToBeOfType
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES,
    data=DATA,
)
def test_failure(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_="NUMBER")
    result = batch_for_datasource.validate(expectation)
    assert not result.success


# Group datasources with case-insensitive type handling
@parameterize_batch_for_data_sources(
    data_source_configs=[
        DatabricksDatasourceTestConfig(),
        PostgreSQLDatasourceTestConfig(),
        SnowflakeDatasourceTestConfig(),
    ],
    data=DATA,
)
def test_case_insensitive_dialects(batch_for_datasource: Batch) -> None:
    dialect_name = batch_for_datasource.data.execution_engine.engine.dialect.name.lower()

    expected_dialects = ["snowflake", "databricks", "postgresql"]
    assert dialect_name in expected_dialects, f"Unexpected dialect: {dialect_name}"

    if dialect_name == "snowflake":
        base_type = "DECIMAL(38, 0)"
    elif dialect_name == "databricks":
        base_type = "INT"
    elif dialect_name == "postgresql":
        base_type = "INTEGER"
    else:
        raise AssertionError(f"Unexpected dialect: {dialect_name}")

    for type_str in [base_type.lower(), base_type.upper(), base_type.capitalize()]:
        expectation = gxe.ExpectColumnValuesToBeOfType(column=INTEGER_COLUMN, type_=type_str)
        result = batch_for_datasource.validate(expectation)
        assert result.success, f"Expected success for type '{type_str}' on dialect '{dialect_name}'"


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param("int", True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_type_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_values_to_be_of_type"
    expectation = gxe.ExpectColumnValuesToBeOfType(
        column=INTEGER_COLUMN,
        type_={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from __future__ import annotations

from great_expectations.compatibility.pyspark import pyspark


class SerializableStructType(dict):
    """Custom type implementing pydantic validation."""

    struct_type: pyspark.sql.types.StructType

    def __init__(
        self,
        fields_or_struct_type: pyspark.sql.types.StructType
        | list[pyspark.sql.types.StructField]
        | None,
    ):
        # Store a copy of the instantiated type as an instance variable
        if isinstance(fields_or_struct_type, pyspark.sql.types.StructType):
            self.struct_type = fields_or_struct_type
        else:
            self.struct_type = pyspark.sql.types.StructType(fields=fields_or_struct_type)

        # Store the serialized version in the keys/values of the instance (parent is dict)
        json_value = self.struct_type.jsonValue()
        super().__init__(**json_value)

    @classmethod
    def validate(
        cls,
        fields_or_struct_type: pyspark.sql.types.StructType
        | list[pyspark.sql.types.StructField]
        | None,
    ):
        """If already StructType then return otherwise try to create a StructType."""
        if isinstance(fields_or_struct_type, pyspark.sql.types.StructType):
            return cls(fields_or_struct_type.fields)
        else:
            return cls(fields_or_struct_type)

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate



from typing import Sequence

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MySQLDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

COL_A = "col_a"
COL_B = "col_b"


DATA = pd.DataFrame(
    {
        COL_A: ["aa", "ab", "ac", None],
        COL_B: ["aa", "bb", "cc", None],
    }
)

SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    PandasDataFrameDatasourceTestConfig(),
    PandasFilesystemCsvDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SparkFilesystemCsvDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


class TestNormalSql:
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegex(column=COL_A, regex="a[x-z]"),
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegex(column=COL_A, regex="[a-z]{99}"),
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegex(column=COL_B, regex="a.", mostly=0.6),
                id="mostly",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
    def test_success(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchRegex,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert result.success

    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegex(column=COL_A, regex="^a[abc]$"),
                id="all_matches",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegex(column=COL_B, regex="a.", mostly=0.7),
                id="mostly_threshold_not_met",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegex(column=COL_A, regex=""), id="empty_regex"
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
    def test_failure(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchRegex,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert not result.success



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Mapping, Optional

import pytest

from great_expectations.compatibility.pydantic import BaseSettings
from great_expectations.compatibility.sqlalchemy import sqltypes
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)
from tests.integration.test_utils.data_source_config.sql import (
    InferrableTypesLookup,
    SQLBatchTestSetup,
)

if TYPE_CHECKING:
    import pandas as pd


class DatabricksDatasourceTestConfig(DataSourceTestConfig):
    @property
    @override
    def label(self) -> str:
        return "databricks"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.databricks

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        return DatabricksBatchTestSetup(
            data=data,
            config=self,
            extra_data=extra_data,
            table_name=self.table_name,
            context=context,
            engine_manager=engine_manager,
        )


class DatabricksBatchTestSetup(SQLBatchTestSetup[DatabricksDatasourceTestConfig]):
    @property
    @override
    def connection_string(self) -> str:
        assert self.schema
        return self._databrics_connection_config.connection_string(self.schema)

    @property
    @override
    def use_schema(self) -> bool:
        return True

    @property
    @override
    def inferrable_types_lookup(self) -> InferrableTypesLookup:
        # databricks requires a length for VARCHAR
        overrides: InferrableTypesLookup = {
            str: sqltypes.VARCHAR(255),
        }
        return super().inferrable_types_lookup | overrides

    @cached_property
    def _databrics_connection_config(self) -> DatabricksConnectionConfig:
        return DatabricksConnectionConfig()  # type: ignore[call-arg]  # retrieves env vars

    @override
    def make_asset(self) -> TableAsset:
        return self.context.data_sources.add_databricks_sql(
            name=self._random_resource_name(),
            connection_string=self.connection_string,
        ).add_table_asset(
            name=self._random_resource_name(),
            table_name=self.table_name,
            schema_name=self.schema,
        )


class DatabricksConnectionConfig(BaseSettings):
    databricks_token: str
    databricks_host: str
    databricks_http_path: str

    def connection_string(self, schema: str) -> str:
        return (
            "databricks://token:"
            f"{self.databricks_token}@{self.databricks_host}:443"
            f"?http_path={self.databricks_http_path}&catalog=ci&schema={schema}"
        )
